import re
import time
import copy
import os
from vaitk import core
import contextlib
from .TextDocumentCursor import TextDocumentCursor
from .LineMetaInfo import LineMetaInfo
from .DocumentMetaInfo import DocumentMetaInfo

EOL='\n'
CHAR_META_INDEX = 0
TEXT_INDEX = 1

class CharMeta:
    LexerToken = "LexerToken"

class TextDocument(core.VObject):
    """
    Represents the contents of a file.
    """

    class MissingFilenameException(Exception): pass

    def __init__(self):
        self._initSignals()

        # New dictionary to store meta info objects. This will outdate the current format.
        self._meta_info = {}

        # Char meta will be outdated, leaving just the text as the contents
        self._contents = [ ({}, EOL) ]

        self.createDocumentMetaInfo("Filename", None)
        self.createDocumentMetaInfo("Modified", False)
        self.createDocumentMetaInfo("LastModified", time.time())
        self.createDocumentMetaInfo("FileType", "Text")

        self._cursors = []

    def __str__(self):
        return self.documentText()

    # Query routines
    def isEmpty(self):
        return len(self._contents) == 1 \
                and len(self._contents[0][TEXT_INDEX]) == 1 \
                and self._contents[0][TEXT_INDEX][0] == EOL

    def isLineEmpty(self, line_number):
        line_index = line_number - 1
        return len(self._contents[line_index][TEXT_INDEX]) == 1 \
                and self._contents[line_index][TEXT_INDEX][0] == EOL

    def lineText(self, line_number):
        self._checkLineNumber(line_number)
        line_index = line_number - 1
        return self._contents[line_index][TEXT_INDEX]

    def hasLine(self, line_number):
        try:
            self._checkLineNumber(line_number)
        except:
            return False
        return True

    def lineLength(self, line_number):
        return len(self.lineText(line_number))

    def documentText(self):
        return "".join([x[TEXT_INDEX] for x in self._contents])

    def filename(self):
        return self.documentMetaInfo("Filename").data()

    def setFilename(self, filename):
        self.documentMetaInfo("Filename").setData(filename)

    def numLines(self):
        return len(self._contents)

    def isModified(self):
        return self.documentMetaInfo("Modified").data()

    ## Meta info routines
    # Document Meta

    def lastModified(self):
        return self.documentMetaInfo("LastModified").data()

    # New interface for document meta info. Will replace the above one
    def createDocumentMetaInfo(self, meta_type, data=None):
        if meta_type in self._meta_info:
            return

        self._meta_info[meta_type] = DocumentMetaInfo(meta_type, self, data)

    def documentMetaInfo(self, meta_type):
        # FIXME currently no safeguard around correct type. Not needed
        # as we will just have metaInfo() in the future
        return self._meta_info[meta_type]
        
    # Line meta

    def createLineMetaInfo(self, meta_type):
        if meta_type in self._meta_info:
            return

        self._meta_info[meta_type] = LineMetaInfo(meta_type, self)

    def lineMetaInfo(self, meta_type):
        # FIXME currently no safeguard around correct type. Not needed
        # as we will just have metaInfo() in the future
        return self._meta_info[meta_type]

    def allLineMetaInfo(self):
        return { k: v for k, v in self._meta_info.items() if isinstance(v, LineMetaInfo) }

    # Char meta

    def charMeta(self, pos):
        self._checkPos(pos)

        line_number, char_number = pos
        line_index = line_number - 1
        char_index = char_number - 1

        how_many = char_number - self.lineLength(line_number)
        char_meta = self._contents[line_index][CHAR_META_INDEX]
        ret = {}
        for key, value in char_meta.items():
            ret[key] = value[char_index:]

        return ret

    def updateCharMeta(self, pos, meta_dict):
        self._checkPos(pos)
        line_number, char_number = pos
        line_index = line_number - 1
        char_index = char_number - 1

        char_meta = self._contents[line_index][CHAR_META_INDEX]
        text = self._contents[line_index][TEXT_INDEX]
        for key, value in meta_dict.items():
            if not key in char_meta:
                char_meta[key] = [None]*len(text)

            char_meta[key][char_index:char_index+len(value)] = value
            char_meta[key] = char_meta[key][0:len(text)]

        self.metaContentChanged.emit()

    def deleteCharMeta(self, pos, how_many, keys):
        line_number, char_number = pos
        line_index = line_number - 1
        char_index = char_number - 1

        char_meta = self._contents[line_index][CHAR_META_INDEX]
        for key in keys:
            try:
                meta_values = char_meta[key]
            except KeyError:
                continue
            meta_values[char_index:char_index+how_many] = None
            char_meta[key] = meta_values

        self.metaContentChanged.emit()

    def wordAt(self, pos, split_func=None):
        """
        find the full word at a given document position.
        Returns a tuple with the full word and the document column where the
        word starts
        """
        self._checkPos(pos)

        if split_func == None:
            split_func = re.compile("(\w+)").finditer

        line_text = self.lineText(pos[0])

        res = list(filter(lambda x: x.start() <= pos[1]-1 < x.end(), split_func(line_text)))

        if len(res) == 1:
            return (res[0].group(0), res[0].start()+1)

        return ('', None)

    def charAt(self, pos):
        """
        find the full word at a given document position.
        Returns a tuple with the full word and the document column where the
        word starts
        """
        self._checkPos(pos)

        line_text = self.lineText(pos[0])

        return line_text[pos[1]-1]
        
    ## Modify document routines
    # Line operations
    def newLineAfter(self, line_number):
        self._checkLineNumber(line_number)
        line_index = line_number - 1

        # Add an EOL if not already there
        self._contents[line_index] = ( self._contents[line_index][CHAR_META_INDEX],
                                       _withEOL(self._contents[line_index][TEXT_INDEX])
                                      )

        self._contents.insert(line_index+1, ({}, EOL))
        self._setModified(True)

        for meta in self.allLineMetaInfo().values():
            meta.addLines(line_number+1, 1)

        for meta in self.allLineMetaInfo().values():
            meta.notifyObservers()

        self.contentChanged.emit()
        self.metaContentChanged.emit()
        self.numLinesChanged.emit()

    def newLine(self, line_number):
        line_index = line_number - 1
        self._contents.insert(line_index, ({}, EOL))
        self._setModified(True)

        for meta in self.allLineMetaInfo().values():
            meta.addLines(line_number, 1)

        for meta in self.allLineMetaInfo().values():
            meta.notifyObservers()

        self.contentChanged.emit()
        self.metaContentChanged.emit()
        self.numLinesChanged.emit()

    def insertLine(self, line_number, text, char_meta=None):
        if not (1 <= line_number <= self.numLines()+1):
            raise IndexError("Invalid insertion line %d" % line_number)

        line_index = line_number - 1
        char_meta = {} if char_meta is None else char_meta
        self._contents.insert(line_index, [char_meta, _withEOL(text)])
        self._setModified(True)

        for meta in self.allLineMetaInfo().values():
            meta.addLines(line_number, 1)

        for meta in self.allLineMetaInfo().values():
            meta.notifyObservers()

        self.contentChanged.emit()
        self.metaContentChanged.emit()
        self.numLinesChanged.emit()

    def insertLines(self, insert_at, text_lines):
        if not (1 <= insert_at <= self.numLines()+1):
            raise IndexError("Invalid insertion line %d" % insert_at)

        insert_at_index = insert_at - 1
        for idx, text in enumerate(text_lines):
            self._contents.insert(insert_at+idx, ( {}, _withEOL(text)))

        for meta in self.allLineMetaInfo().values():
            meta.addLines(insert_at, len(text_lines))

        self._setModified(True)

        for meta in self.allLineMetaInfo().values():
            meta.notifyObservers()
        self.contentChanged.emit()
        self.metaContentChanged.emit()
        self.numLinesChanged.emit()

    def deleteLine(self, line_number):
        self._checkLineNumber(line_number)
        line_index = line_number - 1
        self._contents.pop(line_index)
        self._setModified(True)
        if len(self._contents) == 0:
            self._contents.append(({}, EOL))

        for meta in self.allLineMetaInfo().values():
            meta.deleteLines(line_number, 1)

        for meta in self.allLineMetaInfo().values():
            meta.notifyObservers()

        self.contentChanged.emit()
        self.metaContentChanged.emit()
        self.numLinesChanged.emit()

    def deleteLines(self, from_line, how_many):
        self._checkLineNumber(from_line)
        self._checkLineNumber(from_line+how_many)
        from_line_index = from_line - 1
        self._contents = self._contents[:from_line_index] + self._contents[from_line_index+how_many:]

        if len(self._contents) == 0:
            self._contents.append(({}, EOL))

        self._setModified(True)

        for meta in self.allLineMetaInfo().values():
            meta.deleteLines(from_line, how_many)

        for meta in self.allLineMetaInfo().values():
            meta.notifyObservers()

        self.contentChanged.emit()
        self.metaContentChanged.emit()
        self.numLinesChanged.emit()

    def replaceLine(self, line_number, text, char_meta=None):
        self._checkLineNumber(line_number)

        line_index = line_number - 1
        self._contents.pop(line_index)
        char_meta = {} if char_meta is None else char_meta

        self._contents.insert(line_index, (char_meta, _withEOL(text)))
        self._setModified(True)
        self.contentChanged.emit()
        self.metaContentChanged.emit()

    def breakLine(self, pos):
        self._checkPos(pos)
        line_number, char_number = pos
        line_index = line_number - 1
        char_index = char_number - 1

        current_line_contents = self._contents.pop(line_index)

        orig_char_meta = current_line_contents[CHAR_META_INDEX]
        orig_text      = current_line_contents[TEXT_INDEX]

        above_char_meta = {}
        below_char_meta = {}
        for key, values in orig_char_meta.items():
           above_char_meta[key] = values[:char_index] + [None]
           below_char_meta[key] = values[char_index:]

        above_text = _withEOL(orig_text[:char_index])
        below_text = _withEOL(orig_text[char_index:])

        self._contents.insert(line_index, (below_char_meta,
                                           below_text
                                          )
                            )

        self._contents.insert(line_index, ( above_char_meta,
                                           above_text
                                          )
                            )

        for meta in self.allLineMetaInfo().values():
            meta.addLines(line_number, 1)

        self._setModified(True)
        for meta in self.allLineMetaInfo().values():
            meta.notifyObservers()
        self.contentChanged.emit()
        self.metaContentChanged.emit()
        self.numLinesChanged.emit()

    def joinWithNextLine(self, line_number):
        self._checkLineNumber(line_number)

        if not self.hasLine(line_number+1):
            return

        line_index = line_number - 1

        if self.isLineEmpty(line_number):
            if not self.isEmpty():
                self._contents.pop(line_index)
                for meta in self.allLineMetaInfo().values():
                    meta.deleteLines(line_number, 1)
                self._setModified(True)
                for meta in self.allLineMetaInfo().values():
                    meta.notifyObservers()
                self.contentChanged.emit()
                self.metaContentChanged.emit()
                self.numLinesChanged.emit()
            return

        current_line_contents = self._contents.pop(line_index)
        current_line_char_meta = current_line_contents[CHAR_META_INDEX]
        current_line_text = current_line_contents[TEXT_INDEX]

        next_line_contents = self._contents.pop(line_index)
        next_line_char_meta = next_line_contents[CHAR_META_INDEX]
        next_line_text = next_line_contents[TEXT_INDEX]

        # Merge char meta. Collisions: meta will be merged.
        # [1,1] + [2,2] = [1,1,2,2]

        new_char_meta = {}
        all_keys = set(list(current_line_char_meta.keys()) + list(next_line_char_meta.keys()))
        for key in all_keys:
            current_line_char_values = current_line_char_meta.get(key)
            next_line_char_values = next_line_char_meta.get(key)
            if current_line_char_values is None:
                current_line_char_values = [None] * len(_withoutEOL(current_line_text))
            if next_line_char_values is None:
                next_line_char_values = [None] * len(_withEOL(next_line_text))

            new_char_meta[key] = current_line_char_values + next_line_char_values

        self._contents.insert(line_index, ( new_char_meta,
                                             _withoutEOL(current_line_text) + _withEOL(next_line_text)
                                        )
                             )

        for meta in self.allLineMetaInfo().values():
            meta.deleteLines(line_number+1, 1)

        self._setModified(True)
        for meta in self.allLineMetaInfo().values():
            meta.notifyObservers()

        self.contentChanged.emit()
        self.metaContentChanged.emit()
        self.numLinesChanged.emit()

    # Char operations
    def insertChars(self, pos, string):
        self._checkPos(pos)
        line_number, char_number = pos
        line_index = line_number - 1
        char_index = char_number - 1

        contents = self._contents.pop(line_index)
        text = contents[TEXT_INDEX]

        new_text = text[:char_index] + \
                   string + \
                   text[char_index:]

        char_meta = contents[CHAR_META_INDEX]
        for key, values in char_meta.items():
            char_meta[key] = values[:char_index] + \
                             [None] * len(string) + \
                             values[char_index:]

        self._contents.insert(line_index, ( char_meta,
                                            new_text
                                          )
                                )

        self._setModified(True)
        self.contentChanged.emit()
        self.metaContentChanged.emit()

    def deleteChars(self, pos, how_many):
        """
        Deletes at max how_many characters, starting and including
        position pos. The EOL is never deleted.
        Returns a tuple containing the deleted string and the character
        metainfo for that string.
        """
        self._checkPos(pos)
        if how_many < 0:
            raise ValueError("Negative how_many passed")

        line_number, char_number = pos
        line_index = line_number - 1
        char_index = char_number - 1
        line_length = self.lineLength(line_number)

        #  0 1 2 3 4 5
        # 'h e l l o \n'  line_length = 6
        #   |       |
        #   |       char_index+how_many = 5. how_many = 4
        #   char_index = 1
        if char_index+how_many > line_length-1:
            how_many = line_length-char_index-1

        contents = self._contents.pop(line_index)
        text = contents[TEXT_INDEX]

        new_text = text[:char_index] + text[char_index+how_many:]
        deleted_text = text[char_index:char_index+how_many]

        new_eol_meta = []
        if not _hasEOL(new_text):
            new_eol_meta = [None]

        char_meta = contents[CHAR_META_INDEX]
        deleted_char_meta = {}
        for key, values in char_meta.items():
            char_meta[key] = values[:char_index] + values[char_index+how_many:] + new_eol_meta
            deleted_char_meta[key] = values[char_index:char_index+how_many]

        self._contents.insert(line_index, ( char_meta,
                                            new_text
                                        )
                                )

        self._setModified(True)
        self.contentChanged.emit()
        self.metaContentChanged.emit()
        return (deleted_text, deleted_char_meta)

    def replaceChars(self, pos, how_many, string):
        self._checkPos(pos)
        if how_many < 0:
            raise ValueError("Negative how_many passed")

        line_number, char_number = pos
        line_index = line_number - 1
        char_index = char_number - 1
        line_length = self.lineLength(line_number)

        #  0 1 2 3 4 5
        # 'h e l l o \n'  line_length = 6
        #   |       |
        #   |       char_index+how_many = 5. how_many = 4
        #   char_index = 1
        if char_index+how_many > line_length-1:
            how_many = line_length-char_index-1

        contents = self._contents.pop(line_index)
        text = contents[TEXT_INDEX]

        new_text = text[:char_index] + string + text[char_index+how_many:]
        deleted_text = text[char_index:char_index+how_many]

        new_eol_meta = []
        if not _hasEOL(new_text):
            new_eol_meta = [None]

        char_meta = contents[CHAR_META_INDEX]
        deleted_char_meta = {}
        for key, values in char_meta.items():
            char_meta[key] = values[:char_index] + [None]*len(string) + values[char_index+how_many:] + new_eol_meta
            deleted_char_meta[key] = values[char_index:char_index+how_many]

        self._contents.insert(line_index, ( char_meta,
                                            new_text
                                        )
                                )

        self._setModified(True)
        self.contentChanged.emit()
        self.metaContentChanged.emit()
        return (deleted_text, deleted_char_meta)

    # Input Output

    def open(self, filename):
        contents = []
        with contextlib.closing(open(filename,'r')) as f:
            for textline in f:
                contents.append(({}, _withEOL(textline).replace("\t", " "*4)))

        if len(contents) == 0:
            contents.append(({}, EOL))

        self.documentMetaInfo("Filename").setData(filename)
        self.documentMetaInfo("Modified").setData(False)
        self.documentMetaInfo("LastModified").setData(time.time())
        self._contents = contents
        self._cursors = []

        for meta in self.allLineMetaInfo().values():
            meta.resetLines()

        for meta in self.allLineMetaInfo().values():
            meta.notifyObservers()

        self.contentChanged.emit()
        self.metaContentChanged.emit()
        self.modifiedChanged.emit(False)
        self.filenameChanged.emit(filename)

    def save(self):
        if self.filename() is None:
            raise TextDocument.MissingFilenameException()

        self.saveAs(self.filename())

    def saveAs(self, filename):
        with contextlib.closing(open(filename,'w')) as f:
            f.write(self.documentText())

        filename_changed = (self.documentMetaInfo("Filename").data() != filename)

        self.documentMetaInfo("Filename").setData(filename)
        self._setModified(False)
        if filename_changed:
            self.filenameChanged.emit(filename)
        self.documentSaved.emit(filename)

    # Cursor handling
    def registerCursor(self, cursor):
        self._cursors.append(cursor)

    def createCursor(self):
        return TextDocumentCursor(self)

    def isValidLine(self, line_number):
        return (1 <= line_number <= len(self._contents))

    def isValidPos(self, pos):
        return (self.isValidLine(pos[0]) and (1 <= pos[1] <= self.lineLength(pos[0])))

    # Memento extraction for a line

    def lineMemento(self, line_number):
        memento = [ copy.deepcopy(self._contents[line_number-1]) ]
        meta_info = {}
        for name, meta in self.allLineMetaInfo().items():
            meta_info[name] = meta.memento(line_number)

        memento.append(meta_info)
        return memento

    def insertFromMemento(self, line_number, memento):
        self._contents.insert(line_number-1, copy.deepcopy(memento[0]))
        for name, meta_info in memento[1].items():
            self._meta_info[name].insertFromMemento(line_number, meta_info)

        for meta in self.allLineMetaInfo().values():
            meta.notifyObservers()

        self.contentChanged.emit()
        self.metaContentChanged.emit()
        self.numLinesChanged.emit()

    def replaceFromMemento(self, line_number, memento):
        self._contents[line_number-1] = copy.deepcopy(memento[0])
        for name, meta_info in memento[1].items():
            self._meta_info[name].replaceFromMemento(line_number, meta_info)

        for meta in self.allLineMetaInfo().values():
            meta.notifyObservers()

        self.contentChanged.emit()
        self.metaContentChanged.emit()

    # Private

    def _initSignals(self):
        self.contentChanged = core.VSignal(self)
        self.metaContentChanged = core.VSignal(self)
        self.modifiedChanged = core.VSignal(self)
        self.filenameChanged = core.VSignal(self)
        self.documentSaved = core.VSignal(self)
        self.numLinesChanged = core.VSignal(self)

    def _checkLineNumber(self, line_number):
        if not self.isValidLine(line_number):
            raise IndexError("Out of bound. line_number = %d, len = %d" % (line_number, len(self._contents)))

    def _checkPos(self, pos):
        if not self.isValidPos(pos):
            raise IndexError("Out of bound. pos = %s" % str(pos))

    def _setModified(self, modified):
        if modified:
            self.documentMetaInfo("LastModified").setData(time.time())

        if self.documentMetaInfo("Modified").data() != modified:
            self.documentMetaInfo("Modified").setData(modified)
            self.modifiedChanged.emit(modified)


def _withEOL(text):
    if len(text) == 0 or text[-1] != EOL:
        return text+EOL
    return text

def _withoutEOL(text):
    if len(text) == 0 or text[-1] != EOL:
        return text
    return text[:-1]

def _hasEOL(text):
    return (len(text) != 0 and text[-1] == EOL)



from vaitk import core

class DocumentMetaInfo:
    def __init__(self, meta_type, document, data=None):
        self._meta_type = meta_type
        self._document = document
        self._data = data
        self.contentChanged = core.VSignal(self)

    def setData(self, data):
        if self._data != data:
            self._data = data
            self.notifyObservers()

    def data(self):
        return self._data

    def clear(self):
        if self._data is not None:
            self._data = None
            self.notifyObservers()

    def notifyObservers(self):
        self.contentChanged.emit()

    @property
    def meta_type(self):
        return self._meta_type

    @property
    def document(self):
        return self._document

    def __str__(self):
        return str(self._data)



from vaitk import core
import collections
import copy

# New class to store meta information in a separate
# object, so that we can listen to specific notifications
# from specific meta objects. We need to keep it synchronized
# to the current document though.
class LineMetaInfo:
    def __init__(self, meta_type, document):
        self._meta_type = meta_type
        self._document = document
        self.resetLines()

        self.contentChanged = core.VSignal(self)

    def numLines(self):
        return len(self._data)

    # Meant to be called by document.
    def addLines(self, line_number, how_many):
        for i in range(how_many):
            self._data.insert(line_number-1, None)

    # Meant to be called by document.
    def deleteLines(self, line_number, how_many):
        for i in range(how_many):
            self._data.pop(line_number-1)

    # Meant to be called by document.
    def resetLines(self):
        self._data = [None] * self._document.numLines()

    def setData(self, data, from_line=1):
        # As a method, so we can bind to it via signal/slot. Property with slice needed.
        if not isinstance(data, collections.Iterable) or isinstance(data, str):
            data = [ data ]

        try:
            for idx, d in enumerate(data):
                self._data[from_line-1+idx] = d
        except IndexError:
            pass

        self.notifyObservers()

    def setDataForLines(self, data_dict):
        for k, v in data_dict.items():
            self._data[k-1] = v

        self.notifyObservers()

    def data(self, from_line=None, how_many=None):
        if from_line is not None:
            if how_many is None:
                return self._data[from_line-1]
            else:
                return self._data[from_line-1:from_line-1+how_many]
        else:
            if how_many is None:
                return self._data
            else:
                return self._data[:how_many]

    def notNoneData(self):
        return self.findWhere(lambda x: x is not None)

    def findWhere(self, condition):
        return { i+1: v for i,v in enumerate(self._data) if condition(v)}

    def dataForLines(self, lines):
        return {i: self._data[i-1] for i in lines}

    def clear(self):
        self._data = [None] * self._document.numLines()
        self.notifyObservers()

    def notifyObservers(self):
        self.contentChanged.emit()

    @property
    def meta_type(self):
        return self._meta_type

    @property
    def document(self):
        return self._document

    def memento(self, line):
        return copy.deepcopy(self._data[line-1])

    def insertFromMemento(self, line, memento):
        self._data.insert(line-1, copy.deepcopy(memento))

    def replaceFromMemento(self, line, memento):
        self._data[line-1] = copy.deepcopy(memento)

    def __str__(self):
        return str(self._data)



from .LinterResult import LinterResult
import pyflakes.api

class PyFlakesLinter:
    """Fast, less accurate linter."""
    def __init__(self, document):
        self._document = document

    def runOnce(self):
        reporter = Reporter()
        pyflakes.api.check(self._document.documentText(),
                           self._document.filename(),
                           reporter=reporter)
        return reporter.errors()

class Reporter:
    """Wrap class to get events delivered for our consumption"""
    def __init__(self):
        self._errors = []

    def unexpectedError(self, *args):
        pass

    def syntaxError(self, filename, msg, lineno, offset, text):
        self._errors.append( LinterResult(filename = filename,
                                          line = lineno,
                                          column = offset,
                                          level = "error",
                                          message = msg)
                            )

    def flake(self, msg):
        self._errors.append( LinterResult(filename = msg.filename,
                                          line = msg.lineno,
                                          column = msg.col,
                                          level = "warning",
                                          message = str(msg))
                            )

    def errors(self):
        return self._errors




from .VPoint import VPoint
from .VSize import VSize
#  012345678901234567890
# 0
# 1
# 2  AAAAA
# 3  AAAAA
# 4  AAAAA
# 5       BBB
# 6       BBB
# 7       BBB
# 8
# A = (2,2),(5,3) topleft 2,2 bottomright 6,4

class VRect(object):
    def __init__(self, top_left, size):
        if isinstance(top_left, VPoint):
            self._top_left = top_left
        else:
            self._top_left = VPoint(x=top_left[0], y=top_left[1])

        if isinstance(size, VSize):
            self._size = size
        else:
            self._size = VSize(width=size[0], height=size[1])

    @property
    def size(self):
        return self._size
    @property
    def height(self):
        return self._size.height
    @property
    def width(self):
        return self._size.width

    def isNull(self):
        return (self.width == 0 and self.height == 0)

    def moveTo(self, top_left):
        self._top_left = top_left

    def intersects(self, other):
        return (self.left <= other.right
                and self.right >= other.left
                and self.top <= other.bottom
                and self.bottom >= other.top)

    @property
    def x(self):
        return self.top_left.x
    @property
    def y(self):
        return self.top_left.y
    @property
    def left(self):
        return self.x
    @property
    def right(self):
        return self.left + self.width - 1
    @property
    def top(self):
        return self.y
    @property
    def bottom(self):
        return self.top + self.height - 1
    @property
    def top_left(self):
        return self._top_left
    @property
    def top_right(self):
        return VPoint(self.right, self.top)
    @property
    def bottom_left(self):
        return VPoint(self.left, self.bottom)
    @property
    def bottom_right(self):
        return VPoint(self.right, self.bottom)

    def __str__(self):
        return "VRect(x=%d, y=%d, width=%d, height=%d)" % (self.x, self.y, self.width, self.height)

    class tuple:
        @staticmethod
        def x(rect):
            return rect[0]
        @staticmethod
        def y(rect):
            return rect[1]
        @staticmethod
        def left(rect):
            return rect[0]
        @staticmethod
        def right(rect):
            return rect[0] + rect[2] - 1
        @staticmethod
        def top(rect):
            return rect[1]
        @staticmethod
        def bottom(rect):
            return rect[1] + rect[3] - 1
        @staticmethod
        def topLeft(rect):
            return (rect[0], rect[1])
        @staticmethod
        def topRight(rect):
            return (rect[0] + rect[2] - 1, rect[1])
        @staticmethod
        def bottomLeft(rect):
            return (rect[0], rect[1] + rect[3] - 1)
        @staticmethod
        def bottomRight(rect):
            return (rect[0] + rect[2] - 1, rect[1] + rect[3] - 1)
        @staticmethod
        def size(rect):
            return (rect[2], rect[3])
        @staticmethod
        def height(rect):
            return rect[3]
        @staticmethod
        def width(rect):
            return rect[2]
        @staticmethod
        def isNull(rect):
            return (rect[2] == 0 and rect[3] == 0)
        @staticmethod
        def moveTo(rect, top_left):
            return (top_left[0], top_left[1], rect[2], rect[3])
        @staticmethod
        def intersects(rect, other):
            return (VRect.tuple.left(rect) <= VRect.tuple.right(other)
                    and VRect.tuple.right(rect) >= VRect.tuple.left(other)
                    and VRect.tuple.top(rect) <= VRect.tuple.bottom(other)
                    and VRect.tuple.bottom(rect) >= VRect.tuple.top(other))




from ..core.VObject import VObject

class VSignalSpy(VObject):
    def __init__(self, signal):
        self._signal_params = []
        self._signal = signal
        self._signal.connect(self._signalReceived)

    def _signalReceived(self, *args, **kwargs):
        self._signal_params.append( (args, kwargs) )

    def count(self):
        return len(self._signal_params)

    def lastSignalParams(self):
        return self._signal_params[-1]

    def signalParams(self):
        return self._signal_params




from .. import models

class StatusBarController(object):
    def __init__(self, status_bar):
        self._status_bar = status_bar
        self._buffer = None

    @property
    def buffer(self):
        return self._buffer

    @buffer.setter
    def buffer(self, buffer):
        if buffer is None:
            raise Exception("Cannot set buffer to None")

        if self._buffer is not None:
            self._buffer.document.modifiedChanged.disconnect(self._status_bar.setFileChangedFlag)
            self._buffer.document.filenameChanged.disconnect(self._status_bar.setFilename)
            self._buffer.cursor.positionChanged.disconnect(self._status_bar.setPosition)

        self._buffer = buffer

        # Connect the signals
        self._buffer.document.modifiedChanged.connect(self._status_bar.setFileChangedFlag)
        self._buffer.document.filenameChanged.connect(self._status_bar.setFilename)
        self._buffer.cursor.positionChanged.connect(self._status_bar.setPosition)

        # Update the widget
        self._status_bar.setFilename(self._buffer.document.filename())
        self._status_bar.setFileChangedFlag(self._buffer.document.isModified())
        self._status_bar.setPosition(self._buffer.cursor.pos)



from ..widgets import LineBadge
from ..models import Configuration, Icons
from vaitk import gui

class SideRulerController:
    def __init__(self, side_ruler):
        self._side_ruler = side_ruler


        self._buffer = None
        self._icons = Icons.getCollection(Configuration.get("icons.collection"))

    @property
    def buffer(self):
        return self._buffer

    @buffer.setter
    def buffer(self, buffer):
        if buffer is None:
            raise Exception("Cannot set buffer to None")

        if self._buffer:
            self._buffer.edit_area_model.documentPosAtTopChanged.disconnect(self.updateWidget)
            self._buffer.document.numLinesChanged.disconnect(self.updateWidget)
            self._buffer.document.lineMetaInfo("LinterResult").contentChanged.disconnect(self.updateWidget)
            self._buffer.document.lineMetaInfo("Change").contentChanged.disconnect(self.updateWidget)
            self._buffer.document.lineMetaInfo("Bookmark").contentChanged.disconnect(self.updateWidget)

        self._buffer = buffer

        # Set signals
        self._buffer.edit_area_model.documentPosAtTopChanged.connect(self.updateWidget)
        self._buffer.document.numLinesChanged.connect(self.updateWidget)
        self._buffer.document.lineMetaInfo("LinterResult").contentChanged.connect(self.updateWidget)
        self._buffer.document.lineMetaInfo("Change").contentChanged.connect(self.updateWidget)
        self._buffer.document.lineMetaInfo("Bookmark").contentChanged.connect(self.updateWidget)

        # Refresh
        self.updateWidget()

    def updateWidget(self, *args):
        if self._buffer is None:
            return

        top_pos = self._buffer.edit_area_model.document_pos_at_top
        self._side_ruler.setTopLine(top_pos[0])
        self._side_ruler.setNumLines(self._buffer.document.numLines())

        badges = {}
        needed_lines = self._side_ruler.visibleLineNumbers()
        changed_data = self._buffer.document.lineMetaInfo("Change").dataForLines(needed_lines)

        for line, change in changed_data.items():
            if change == "added":
                badges[line] = LineBadge(marker=self._icons["SideRuler.added"],
                                         fg_color=gui.VGlobalColor.green,
                                         bg_color=None)
            elif change == "modified":
                badges[line] = LineBadge(marker=self._icons["SideRuler.modified"],
                                         fg_color=gui.VGlobalColor.magenta,
                                         bg_color=None)
            elif change == "deletion_before":
                badges[line] = LineBadge(marker=self._icons["SideRuler.deletion_before"],
                                         fg_color=gui.VGlobalColor.red,
                                         bg_color=None)
            elif change == "deletion_after":
                badges[line] = LineBadge(marker=self._icons["SideRuler.deletion_after"],
                                         fg_color=gui.VGlobalColor.red,
                                         bg_color=None)

        bookmarks = self._buffer.document.lineMetaInfo("Bookmark").dataForLines(needed_lines)

        for line, mark in bookmarks.items():
            if mark is None:
                continue

            badges[line] = LineBadge(marker=self._icons["SideRuler.bookmarks"][ord(mark)-ord('a')],
                                  fg_color=gui.VGlobalColor.yellow,
                                  bg_color=None,
                        )

        lint_data = self._buffer.document.lineMetaInfo("LinterResult").dataForLines(needed_lines)

        for line, lint in lint_data.items():
            if lint is None:
                continue

            if lint.level == "error":
                badges[line] = LineBadge(marker=self._icons["SideRuler.error"],
                                  fg_color=gui.VGlobalColor.red,
                                  bg_color=None
                        )
            elif lint.level == "warning":
                badges[line] = LineBadge(marker=self._icons["SideRuler.warning"],
                                  fg_color=gui.VGlobalColor.yellow,
                                  bg_color=None
                        )
            elif lint.level == "info":
                badges[line] = LineBadge(marker=self._icons["SideRuler.info"],
                                  fg_color=gui.VGlobalColor.cyan,
                                  bg_color=None
                                )

        self._side_ruler.setBadges(badges)
        self._side_ruler.update()





from ... import core
from ..VWidget import VWidget
from ..VPainter import VPainter
from ..VPalette import VPalette

class VProgressBar(VWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self._text = ""
        self._minimum = 0
        self._maximum = 100
        self._value = 0
    def paintEvent(self, event):
        painter = VPainter(self)
        w, h = self.size()
        fg_color = self.palette().color(VPalette.ColorGroup.Active,
                                        VPalette.ColorRole.WindowText)
        bg_color = self.palette().color(VPalette.ColorGroup.Active,
                                        VPalette.ColorRole.Window)

        final_text = ""
        if len(self._text) > 0:
            final_text += (self._text+ " ")

        percentage = int(100 * (self.value() - self.minimum())/(self.maximum() - self.minimum()))
        percentage_text = ("%d" % percentage) + "%"
        bar_total_length = w - len(final_text) - 2
        if bar_total_length < 5:
            bar_text = percentage_text.rjust(4)
        else:
            bar_fill_length = int(bar_total_length * percentage / 100)
            bar_text = ("="*bar_fill_length) + (' '*(bar_total_length-bar_fill_length))
            bar_text = bar_text[:int(bar_total_length/2)-1] + \
                       percentage_text + \
                       bar_text[int(bar_total_length/2)-1+len(percentage_text):]

        final_text += '[' + bar_text + ']'
        painter.drawText( (0, 0), final_text, fg_color, bg_color)

    def minimumSize(self):
        if len(self._text) > 0:
            width = len(self._text)+len(" [100%]")
        else:
            width = len("[100%]")

        return (width, 1)

    def setValue(self, value):
        if self._value != value and (self._minimum < value < self._maximum):
            self._value = value
            self.update()

    def minimum(self):
        return self._minimum
    def maximum(self):
        return self._maximum
    def setMinimum(self, minimum):
        if self._minimum == minimum:
            return

        self._minimum = minimum
        if self._minimum > self._maximum:
            self._maximum = self._minimum

        if not (self._minimum < self._value < self._maximum):
            self.reset()
        else:
            self.update()

    def setMaximum(self, maximum):
        if self._maximum == maximum:
            return

        self._maximum = maximum
        if self._maximum < self._minimum:
            self._minimum = self._maximum

        if not (self._minimum < self._value < self._maximum):
            self.reset()
        else:
            self.update()

    def reset(self):
        self._value = self._minimum
        self.update()

    def setRange(self, minimum, maximum):
        if self._minimum == minimum and self._maximum == maximum:
            return

        self._minimum = minimum
        self._maximum = maximum

        if self._maximum < self._minimum:
            self._maximum = self._minimum

        if not (self._minimum < self._value < self._maximum):
            self.reset()
        else:
            self.update()

    def setText(self, text):
        self._text = text

    def text(self):
        return self._text

    def value(self):
        return self._value



class VSignal(object):
    def __init__(self, sender):
        self._sender = sender
        self._slots = []
        self._enabled = True

    def connect(self, target):
        if isinstance(target, VSignal):
            slot = target.emit
        else:
            slot = target

        if target not in self._slots:
            self._slots.append(slot)

    def disconnect(self, target):
        if target in self._slots:
            self._slots.remove(target)

    def emit(self, *args, **kwargs):
        if not self._enabled:
            return

        for slot in self._slots:
            slot(*args, **kwargs)

    def setEnabled(self, enabled):
        self._enabled = enabled



from .BufferCommand import BufferCommand
from .CommandResult import CommandResult

class ReplaceSingleCharCommand(BufferCommand):
    def __init__(self, buffer, char):
        super().__init__(buffer)
        self._char = char

    def execute(self):
        document = self._buffer.document
        cursor = self._buffer.cursor
        
        if self.savedCursorPos() is None:
            self.saveCursorPos()
        
        pos = self.savedCursorPos()
        cursor.toPos(pos)
         
        self.saveLineMemento(pos[0], BufferCommand.MEMENTO_REPLACE)

        meta_info = document.lineMetaInfo("Change")
        if meta_info.data(pos[0]) is None:
            meta_info.setData('modified', pos[0])

        deleted = document.replaceChars(pos, 1, self._char)
        return CommandResult(True, deleted)




import os
from vaitk import core
from .Buffer import Buffer

class BufferList(core.VObject):
    def __init__(self):
        self._buffers = [Buffer()]
        self._current = self._buffers[0]

        self.currentBufferChanged = core.VSignal(self)

    @property
    def buffers(self):
        return self._buffers

    @property
    def current(self):
        return self._current

    def add(self, buffer):
        self._buffers.append(buffer)
        return buffer

    def bufferForFilename(self, path):
        """
        Given a path, finds the currently open buffer that is associated to that path.
        Returns None if no such buffer exists
        """
        for buffer in self._buffers:
            if buffer.document.filename() is None:
                continue


            try:
                if os.path.samefile(
                        os.path.abspath(
                            os.path.realpath(
                                buffer.document.filename()
                            )
                        ), path):
                    return buffer
            except:
                pass

            if path == buffer.document.filename():
                return buffer

        return None

    def select(self, buffer):
        if buffer not in self._buffers:
            raise Exception("Buffer not in BufferList")

        if buffer != self._current:
            old_buffer = self._current
            self._current = buffer
            self.currentBufferChanged.emit(old_buffer, self._current)

        return self._current

    def addAndSelect(self, buffer):
        self.add(buffer)
        return self.select(buffer)

    def replaceAndSelect(self, old_buffer, new_buffer):
        if old_buffer not in self._buffers:
            raise Exception("Buffer not in BufferList")

        index = self._bufferIndex(old_buffer)
        self._buffers.remove(old_buffer)
        self._buffers.insert(index, new_buffer)
        return self.select(new_buffer)

    def selectNext(self):
        if self.current is None:
            return
        index = (self._bufferIndex(self.current) + 1) % len(self._buffers)
        return self.select(self._buffers[index])

    def selectPrev(self):
        if self.current is None:
            return
        index = (self._bufferIndex(self.current) - 1) % len(self._buffers)
        return self.select(self._buffers[index])

    # Private

    def _bufferIndex(self, buffer):
        return self._buffers.index(buffer)




from vaitk import gui
import math
from ..models import Icons
from ..models import Configuration
from collections import namedtuple

LineBadge = namedtuple('LineBadge', ["marker", "fg_color", "bg_color"])

class SideRuler(gui.VWidget):
    def __init__(self, parent):
        super().__init__(parent)
        self._num_lines = 1
        self._top_line = 1
        self._skip_intervals = []
        self._badges = {}
        self._icons = Icons.getCollection(Configuration.get("icons.collection"))
        self.setColors(gui.VGlobalColor.nameToColor(Configuration.get("colors.side_ruler.fg")),
                       gui.VGlobalColor.nameToColor(Configuration.get("colors.side_ruler.bg"))
                       )

    def paintEvent(self, event):
        w, h = self.size()
        current_fg, current_bg = self.currentColors()
        painter = gui.VPainter(self)
        painter.erase()
        num_digits = self._lineNumberWidth()
        entries = _computeLineValues(self._top_line, h, self._skip_intervals)
        for i, current in enumerate(entries):
            painter.fg_color = current_fg
            painter.bg_color = current_bg

            if current > self._num_lines:
                painter.drawText( (0, i), self._icons["SideRuler.unexistent_line"].ljust(num_digits).ljust(w-1))
                painter.drawText( (w-1, i), self._icons["SideRuler.border"])
                continue


            painter.drawText((0, i),
                                (str(current).rjust(num_digits).ljust(w-2)
                                + " "
                                + self._icons["SideRuler.border"]))

            badge = self._badges.get(current)
            if badge is not None:
                badge_mark = badge.marker
                painter.fg_color = badge.fg_color
                painter.bg_color = badge.bg_color
                painter.drawText( (w-2, i), badge_mark)

    def setNumLines(self, num_lines):
        """Sets the total number of lines the document has"""
        self._num_lines = num_lines
        self.update()

    def setTopLine(self, top_line):
        """Sets the current top line that is displayed"""
        self._top_line = top_line
        self.update()

    def minimumSize(self):
        return (self._lineNumberWidth(), 1)

    def setBadges(self, badges_dict):
        """Set badges at specified lines, overriding the current ones."""
        self._badges.clear()
        self._badges.update(badges_dict)
        self.update()

    def removeBadges(self, lines):
        """Clear specific badge lines"""
        if line in self._badges:
            del self._badges[line]
            self.update()

    def badges(self, lines):
        return [self._badges.get(line) for line in lines]

    def visibleLineNumbers(self):
        w, h = self.size()
        line_values = _computeLineValues(self._top_line, h, self._skip_intervals)
        return [l for l in line_values if l <= self._num_lines]

    # Private

    def _lineNumberWidth(self):
        return int(math.log10(self._num_lines))+1



def _computeLineValues(start, how_many, skip):
    result = []
    current = start
    for i in range(how_many):
        for interval in skip:
            begin, end = interval
            if begin < current < end:
                current = end
        result.append(current)
        current += 1
    return result




from .VObject import VObject
from .VPoint import VPoint
from .VRect import VRect
from .VSize import VSize
from .VCoreApplication import VCoreApplication
from .VTimer import VTimer
from .VSignal import VSignal
from .VEvent import VEvent
from .VTimerEvent import VTimerEvent
from ..consts import Index




from .BufferCommand import BufferCommand
from .CommandResult import CommandResult
from .JoinWithNextLineCommand import JoinWithNextLineCommand

TAB_SPACE = 4

class DeleteSingleCharCommand(BufferCommand):
    def execute(self):
        document = self._document
        cursor = self._cursor

        if self.savedCursorPos() is None:
            self.saveCursorPos()
        
        pos = self.savedCursorPos()
        cursor.toPos(pos)

        if pos == (1,1):
            return CommandResult(success=False, info=None)

        if pos[1] == 1:
            cursor.toPos( (pos[0]-1, document.lineLength(pos[0]-1)) )
            command = JoinWithNextLineCommand(self._buffer)
            result = command.execute()
            if result.success:
                self._sub_command = command
                return CommandResult(True, '\n')
            else:
                return CommandResult(False, None)

        self.saveLineMemento(pos[0], BufferCommand.MEMENTO_REPLACE)

        line_meta = document.lineMetaInfo("Change")
        changed = line_meta.data(pos[0])
        if changed is None:
            line_meta.setData("modified", pos[0])

        # Check if we can remove a tab. These are the tab positions:
        # 123456789
        # T   T   T

        # Get the tab position we can go to, and how many character we need to remove
        # to get there. If we are exactly on a tab position, remove a tab from the last
        # available tab position and add a full tab spacing to potentially remove.
        # FIXME this code's math is ugly. maybe we can simplify it?
        last_tab_pos = 1+int((pos[1]-1)/TAB_SPACE)*TAB_SPACE
        how_many = pos[1]-last_tab_pos
        if last_tab_pos > 1 and how_many == 0:
            how_many = TAB_SPACE
            last_tab_pos = last_tab_pos - TAB_SPACE

        # Check if the candidate removal is only spaces. If yes, then we can remove
        # the whole bunch, otherwise revert back to single char deletion.
        text = document.lineText(pos[0])
        if len(text[last_tab_pos-1:last_tab_pos-1+how_many].strip(' ')) != 0:
            how_many = 1

        deleted = document.deleteChars( (pos[0], pos[1]-how_many), how_many)
        cursor.toPos((pos[0], pos[1]-how_many))

        return CommandResult(success=True, info=deleted)




import contextlib

def report(saved_files):
    print("Apologies. Vai has crashed.")
    print("---------------------------")
    with contextlib.closing(open("vai_crashreport.out")) as f:
        print(f.read())

    print("---------------------------")
    if len(saved_files) != 0:
        print("Your buffers have been dumped to the following files")
        print("")
        for f in saved_files:
            print("  "+str(f))
    print("")
    print("The traceback has been saved in vai_crashreport.out")


def yes():
    y_or_n = input("[y/N]> ")
    if y_or_n.strip().lower() in ['y','yes']:
        return True
    return False





from .BufferCommand import BufferCommand
from .CommandResult import CommandResult

class InsertLineCommand(BufferCommand):
    def __init__(self, buffer, text):
        super().__init__(buffer)
        self._text = text

    def execute(self):
        document = self._document
        cursor = self._cursor

        if self.savedCursorPos() is None:
            self.saveCursorPos()

        pos = self.savedCursorPos()

        document.insertLine(pos[0], self._text)
        document.lineMetaInfo("Change").setData("added", pos[0])
        cursor.toCharFirstNonBlankForLine(pos[0])
        return CommandResult(True, None)

    def undo(self):
        self._document.deleteLine(self._saved_cursor_pos[0])
        self.restoreCursorPos()





from .BreakLineCommand import BreakLineCommand
from .BufferCommand import BufferCommand
from .CommandResult import CommandResult
from .DeleteLineAtCursorCommand import DeleteLineAtCursorCommand
from .DeleteSingleCharAfterCommand import DeleteSingleCharAfterCommand
from .DeleteSingleCharCommand import DeleteSingleCharCommand
from .DeleteToEndOfLineCommand import DeleteToEndOfLineCommand
from .DeleteToEndOfWordCommand import DeleteToEndOfWordCommand
from .InsertLineAfterCommand import InsertLineAfterCommand
from .InsertLineCommand import InsertLineCommand
from .InsertStringCommand import InsertStringCommand
from .JoinWithNextLineCommand import JoinWithNextLineCommand
from .NewLineAfterCommand import NewLineAfterCommand
from .NewLineCommand import NewLineCommand
from .ReplaceSingleCharCommand import ReplaceSingleCharCommand
from .InsertFileCommand import InsertFileCommand
from .IndentCommand import IndentCommand
from .DedentCommand import DedentCommand



'''
import tempfile
import contextlib
import io
import logging
import subprocess
import os
from ..AsyncDocumentProcessor import AsyncDocumentProcessor

class PyLintLinter(AsyncDocumentProcessor):
    """
    Slow, but very accurate  Linter. We need async processing for this one to happen.
    For now it is unused.
    """

    def __init__(self, document, editor):
        super().__init__(document)
        self._editor = editor

    def run(self):
        env = os.environ
        env['LC_ALL']='en_US.UTF-8'
        env['LANG']='en_US.UTF-8'
        stdout = tempfile.NamedTemporaryFile(delete=False)
        stderr = tempfile.NamedTemporaryFile(delete=False)
        proc = subprocess.Popen(['pylint',
                           document.filename(),
                           "--msg-template='%s:{C}:{line}:{column}:{msg_id}:{obj}:{msg}'" % document.filename(),
                           '-r'
                           'n',
                           ],
                           stdout = stdout,
                           stderr = stderr,
                           env=env,
                           universal_newlines=True
                           )
        proc.wait()
        stdout.close()
        stderr.close()
        with contextlib.closing(open(stdout.name)) as f:
            lint_results = f.readlines()

        #with contextlib.closing(open(stderr.name)) as f:
        #    lint_results = f.readlines()

        result = []
        for line in lint_results:
            if line.startswith(document.filename()):
                info = line.split(':')
                info = LinterResult(filename=info[0],
                                    level=info[1],
                                    line=int(info[2]),
                                    column=int(info[3]),
                                    message=info[6].strip()
                                   )
                result.append(info)
        return result

'''



"""
Module that centralizes specific paths we have an interest on.
"""

import os

def configFile():
    """
    Returns the default configuration path in agreement with XDG rules
    """
    config_dir = configDir()
    return os.path.join(config_dir, 'vairc')

def configDir():
    """
    Returns the default configuration dir in agreement with XDG rules
    """
    home = os.path.expanduser('~')
    xdg_config_home = os.environ.get('XDG_CONFIG_HOME') or \
                      os.path.join(home, '.config')
    config_dir = os.path.join(xdg_config_home, 'vai')
    if not os.path.isdir(config_dir):
        os.makedirs(config_dir)
    return config_dir

def stateDir():
    home = os.path.expanduser('~')
    xdg_state_home = os.environ.get('XDG_STATE_HOME') or \
                     os.path.join(home, '.local', 'state')
    state_dir = os.path.join(xdg_state_home, 'vai')
    if not os.path.isdir(state_dir):
        os.makedirs(state_dir)
    return state_dir

def stateFile():
    state_dir = stateDir()
    return os.path.join(state_dir, 'vaistate')

def pluginsDir(scope, category):
    if scope == "system":
        base_path = os.path.join( os.path.dirname(os.path.abspath(__file__)), 'plugins')
    elif scope == "user":
        base_path = os.path.join(os.path.expanduser('~'),
                            '.local',
                            'share',
                            'vai',
                            'plugins'
                           )
    else:
        raise Exception("Invalid plugin dir scope")

    if category not in ["syntaxcolors", "commands"]:
        raise Exception("Invalid plugin dir category")

    path = os.path.join(base_path, category)

    if not os.path.isdir(path):
        os.makedirs(path)

    return path



from .VCoreApplication import VCoreApplication
from .VObject import VObject
from .VSignal import VSignal
from . import VTimerEvent
import time
import threading

class _TimerThread(threading.Thread):
    def __init__(self, timeout, single_shot, callback):
        super(_TimerThread, self).__init__()
        self.daemon = True
        self._timeout = timeout
        self._single_shot = single_shot
        self._callback = callback
        self.stop = threading.Event()

    def run(self):
        while True:
            time.sleep(self._timeout/1000.0)
            if self.stop.is_set():
                break
            self._callback()
            if self._single_shot:
                break

class VTimer(VObject):
    def __init__(self):
        super().__init__()
        self._interval = None
        self._single_shot = False
        self.timeout = VSignal(self)
        self._thread = None
        VCoreApplication.vApp.addTimer(self)

    def start(self):
        if self._thread is not None:
            return
        if self._interval is None:
            return
        self._thread = _TimerThread(self._interval, self._single_shot, self._timeout)
        self._thread.start()

    def _timeout(self):
        VCoreApplication.vApp.postEvent(self, VTimerEvent.VTimerEvent())

    def setSingleShot(self, single_shot):
        self._single_shot = single_shot

    def setInterval(self, interval):
        self._interval = interval

    def stop(self):
        if self._thread:
            self._thread.stop.set()
        self._thread = None

    def isRunning(self):
        return self._thread is not None

    def timerEvent(self, event):
        if isinstance(event, VTimerEvent.VTimerEvent):
            self.timeout.emit()

    @staticmethod
    def singleShot(timeout, callback):
        timer = VTimer()
        timer.setInterval(timeout)
        timer.setSingleShot(True)
        timer.timeout.connect(callback)
        timer.start()





class SymbolLookupDb:
    """
    Singleton class. Contains the database storing the various symbols in the current open codes.
    At the moment, this is filled up by the Lexer.
    Implemented as a simple trie. Nothing fancy.
    """

    _db = {}

    @classmethod
    def clear(cls):
        """
        Completely clears the database
        """
        cls._db = {}

    @classmethod
    def add(cls, word):
        """
        Adds a word to the database
        """
        _add(cls._db, word)

    @classmethod
    def lookup(cls, prefix):
        """
        Given a prefix, looks up all entries having that prefix.
        Returns a list of all postfixes, or an empty list if
        nothing is available.
        """
        d = _walkDown(cls._db, prefix)
        if d is None:
            return []

        ret = _composePostfix(d)
        return ret

# Helper routines for the above class

def _add(d, word):
    """
    Recursive function to add a word to the trie
    """
    if len(word) == 0:
        d[''] = None
        return
    if word[0] not in d:
        d[word[0]] = {}

    _add(d[word[0]], word[1:])

def _walkDown(d, prefix):
    """
    Recursive. Performs traversal of the trie given a prefix
    """
    if len(prefix) == 1:
        if prefix[0] not in d:
            return None
        return d[prefix[0]]

    try:
        return _walkDown(d[prefix[0]], prefix[1:])
    except KeyError:
        return None

def _composePostfix(d, tab=0):
    """
    Recursive function. Builds the postfixes list
    from a given trie branch.
    """

    ret = []
    if d is None:
        return ['']

    for k, v in d.items():
        for postfix in _composePostfix(v, tab+1):
            ret.append(k+postfix)

    return ret



class VPoint(object):
    def __init__(self, x, y):
        self._pos = (x, y)

    def __iter__(self):
        return iter(self._pos)

    @property
    def x(self):
        return self._pos[0]

    @property
    def y(self):
        return self._pos[1]

    def __add__(self, other):
        return VPoint(*VPoint.tuple.add(self._pos, other._pos))

    def __sub__(self, other):
        return VPoint(*VPoint.tuple.sub(self._pos, other._pos))

    def __str__(self):
        return "VPoint(x=%d, y=%d)" % self._pos

    class tuple:

        @staticmethod
        def x(p1):
            return p1[0]

        @staticmethod
        def y(p1):
            return p1[1]

        @staticmethod
        def add(p1, p2):
            return (p1[0]+p2[0], p1[1]+p2[1])

        @staticmethod
        def sub(p1, p2):
            return (p1[0]-p2[0], p1[1]-p2[1])



from .BufferCommand import BufferCommand
from .CommandResult import CommandResult

class InsertStringCommand(BufferCommand):
    """
    Insert a text string in the document at the current cursor position.
    """
    def __init__(self, buffer, text):
        super().__init__(buffer)
        self._text = text

    def execute(self):
        cursor = self._cursor
        document = self._document

        if self.savedCursorPos() is None:
            self.saveCursorPos()

        pos = self.savedCursorPos()

        line_meta = document.lineMetaInfo("Change")
        changed = line_meta.data(pos[0])

        self.saveLineMemento(pos[0], BufferCommand.MEMENTO_REPLACE)

        if changed is None:
            line_meta.setData("modified", pos[0])

        document.insertChars(pos, self._text)
        cursor.toPos( (pos[0], pos[1]+len(self._text)) )

        return CommandResult(success=True, info=None)




from .VLabel import VLabel
from ..VPainter import VPainter
from ..VPalette import VPalette
from ..VPalette import VPalette

class VToolTip(VLabel):
    _instance = None
    @classmethod
    def showText(cls, pos, text):
        """
        Shows the tooltip text at screen position pos.
        Only one tooltip is allowed. If another tooltip is already present,
        it will be moved and the text changed.
        """
        if cls._instance is None:
            cls._instance = VToolTip(text, parent=None)
        cls._instance.setText(text)
        cls._instance.resize((len(text),1))
        cls._instance.move(pos)
        cls._instance.show()

    @classmethod
    def hideText(cls):
        """Hides the tooltip if present"""
        if cls._instance is not None:
            VLabel.hide(cls._instance)

    def paintEvent(self, event):
        painter = VPainter(self)
        w, h = self.size()
        painter.fg_color = self.palette().color(VPalette.ColorGroup.Active,
                                        VPalette.ColorRole.ToolTipText)
        painter.bg_color = self.palette().color(VPalette.ColorGroup.Active,
                                        VPalette.ColorRole.ToolTipBase)
        painter.drawText( (0, 0), self._label)



from .. import FocusPolicy
from .. import core
from . import events
from .VPalette import VPalette
from .VScreen import VScreen
from .events import VFocusEvent
import threading
import queue
import logging
import time

class _VExceptionEvent:
    def __init__(self, exception):
        self.exception = exception

class _KeyEventThread(threading.Thread):
    """
    Separate thread class to handle keyboard events
    extracted from ncurses.
    It communicates to the main thread in three ways:
    - with a queue, where it puts all the key events
    - and with an event flag, which is set when key events are available
    - with the stop_thread flag, which stops the loop

    the reason why we use the event_available_flag is because we have two
    queues, one for the key events, and the other for the other events (timer etc).
    We can't check the queues without altering them, so we need a flag to communicate
    when either of the queues has something to fetch. (to be verified, I remember
    I had to devise this solution due to limits of the queue object)

    Important thing (for future resolution): the stop_event is pretty useless
    since the thread stay stopped in getKeyCode and can't check the flag until
    it leaves. Either I need to add a timer, or come up with a better solution.
    In any case, the self.daemon flag should solve the problem, since the main
    thread is free to quit even if the secondary daemon thread is still running.
    """
    def __init__(self, screen, key_event_queue, event_available_flag):
        super().__init__()
        self.daemon = True
        self.stop_event = threading.Event()
        self.exception = None

        self._screen = screen
        self._key_event_queue = key_event_queue
        self.throttle = threading.Event()
        self._event_available_flag = event_available_flag


    def run(self):
        """
        Runs on the separate thread. Fetches key events from ncurses,
        converts them into vaitk key events, and post them into the key
        event queue for further processing from the main thread.
        """

        while not self.stop_event.is_set():
            last_event = (None, time.time())
            try:
                c = self._screen.getKeyCode()

#                if last_event[0] == c and time.time()-last_event[1] < 0.04:
#                    continue
                last_event = (c, time.time())

                event = events.VKeyEvent.fromNativeKeyCode(c)
                if event is not None:
                    self._key_event_queue.put(event)
                    self._event_available_flag.set()

                    self.throttle.wait()
                    self.throttle.clear()

            except Exception as e:
                event = _VExceptionEvent(e)
                self._key_event_queue.put(event)
                self._event_available_flag.set()

    def registerTimer(self, timer):
        pass

class VApplication(core.VCoreApplication):
    debug=True
    def __init__(self, argv, screen=None):
        from . import VWidget
        super().__init__(argv)
        if screen is not None:
            self._screen = screen
        else:
            self._screen = VScreen()

        self._root_widget = VWidget()
        self._focus_widget = None
        self._palette = self.defaultPalette()
        self._event_available_flag = threading.Event()
        self._event_queue = queue.Queue()
        self._key_event_queue = queue.Queue()
        self._key_event_thread = _KeyEventThread(self._screen, self._key_event_queue, self._event_available_flag)
        self._delete_later_queue = []
        self._exit_flag = False

        # Signals.
        self.lastWindowClosed = core.VSignal(self)
        self.focusChanged = core.VSignal(self)

    def exec_(self):
        self._root_widget.show()
        self.processEvents(True)
        self._key_event_thread.start()
        while self._exit_flag != True:
            self.logger.info("Waiting for events")
            self._event_available_flag.wait()
            self._event_available_flag.clear()
            self.logger.info("Event available")
            self.processEvents(True)
            self._key_event_thread.throttle.set()

        self._exitCleanup()

    def processEvents(self, native=False):
        self.logger.info("++++---- %s processing events ---+++++" % ("Native" if native else "Forced"))
        self._processKeyEvents()
        self._processRemainingEvents()
        self._sendPaintEvents()
        self._deleteScheduled()
        self.logger.info("===================================")
        self._screen.refresh()

    def postEvent(self, receiver, event):
        self.logger.info(" <posted " + str(receiver) + " " + str(event))
        self._event_queue.put((receiver, event))
        self._event_available_flag.set()

    def exit(self):
        self._exit_flag = True

    def addTopLevelWidget(self, widget):
        self._root_widget.addChild(widget)

    def deleteLater(self, widget):
        self.logger.info("Added widget %s to deleteLater queue" % str(widget))
        self._delete_later_queue.append(widget)

    def screen(self):
        return self._screen

    def focusWidget(self):
        return self._focus_widget

    def setFocusWidget(self, widget):
        if self._focus_widget is widget:
            return

        self.logger.info("Setting focus on widget %s." % widget)

        if self._focus_widget is not None:
            self.logger.info("Focus out on widget %s." % self._focus_widget)
            VApplication.vApp.postEvent(self._focus_widget, VFocusEvent(core.VEvent.EventType.FocusOut))

        self._focus_widget = None
        if widget is not None:
            if widget.focusPolicy() == FocusPolicy.NoFocus:
                self.logger.info("Focus not accepted on widget %s due to its focus policy." % self._focus_widget)
                return

            self._focus_widget = widget
            VApplication.vApp.postEvent(self._focus_widget, VFocusEvent(core.VEvent.EventType.FocusIn))

    def defaultPalette(self):
        palette = VPalette()
        palette.setDefaults()
        return palette

    def palette(self):
        return self._palette

    def rootWidget(self):
        try:
            return self._root_widget
        except AttributeError:
            self._root_widget = None
            return self._root_widget

    def resetScreen(self):
        self._screen.reset()

    def topLevelWidgets(self):
        raise NotImplementedError()

    def allWidgets(self):
        raise NotImplementedError()

    def activeWindow(self):
        raise NotImplementedError()

    def setActiveWindow(self, window):
        raise NotImplementedError()

    def closeAllWindows(self):
        raise NotImplementedError()

    def event(self, event):
        raise NotImplementedError()

    def clipboard(self):
        raise NotImplementedError()

    def keyboardModifiers(self):
        raise NotImplementedError()

    def notify(self):
        raise NotImplementedError()

    # Private

    def _hideScheduled(self):
        self.logger.info("Widget scheduled for deletion: %s" % str(self._delete_later_queue))
        for w in self._delete_later_queue:
            self.logger.info("Posting hide events for deleted widget %s" % str(w))
            w.hide()

    def _deleteScheduled(self):
        for w in self._delete_later_queue:
            w.parent().removeChild(w)

        self._delete_later_queue.clear()

    def _processKeyEvents(self):
        while True:
            self.logger.info("key queue %d" % self._key_event_queue.qsize())
            try:
                event = self._key_event_queue.get_nowait()
            except queue.Empty:
                event = None

            if event is None:
                return

            if isinstance(event, events.VKeyEvent):
                self._processSingleKeyEvent(event)
            elif isinstance(event, _VExceptionEvent):
                raise event.exception

    def _processSingleKeyEvent(self, key_event):
        self.logger.info("Key event %d %x" % (key_event.key(), key_event.modifiers()))

        if not self.focusWidget():
            self.logger.info("Key event ignored. No widget has focus.")
            return

        key_event.setAccepted(False)
        for widget in self.focusWidget().traverseToRoot():
            self.logger.info("KeyEvent attempting delivery to "+str(widget))
            stop_event = False
            for event_filter in reversed(widget.installedEventFilters()):
                stop_event = stop_event | event_filter.eventFilter(key_event)
                if key_event.isAccepted():
                    self.logger.info("KeyEvent accepted by filter "+str(event_filter))
                    return

            if not stop_event:
                self.logger.info("KeyEvent not stopped. Sending to widget "+str(widget))
                widget.keyEvent(key_event)

                if key_event.isAccepted():
                    self.logger.info("KeyEvent accepted by "+str(widget))
                    return

    def _processRemainingEvents(self):
        previous_data = None
        while True:
            try:
                receiver, event = self._event_queue.get_nowait()
            except queue.Empty:
                receiver, event = None, None

            if event is None:
                break

            if previous_data is not None:
                prev_receiver, prev_event = previous_data
                if event.eventType() == prev_event.eventType() and receiver == prev_receiver:
                   continue

            self.logger.info("Data queue %d. Processing %s -> %s." % (self._event_queue.qsize(), str(event), str(receiver)))
            receiver.event(event)
            previous_data = (receiver, event)

            #self._stop_flag.append(1)
        # Check if screen was re-sized (True or False)
        #x,y = self._screen.size()
        #resize = curses.is_term_resized(y, x)

        # Action in loop if resize is True:
        #if resize is True:
            #x, y = self._screen.size()
            #curses.resizeterm(y, x)
            #self.renderWidgets()

    def _sendPaintEvents(self):
        for w in self.rootWidget().depthFirstFullTree():
            if w.needsUpdate():
                for w2 in w.depthFirstRightTree():
                    if core.VRect.tuple.intersects(w.absoluteRect(),w2.absoluteRect()):
                        w2.update()

        for w in self.rootWidget().depthFirstFullTree():
            if w.needsUpdate():
                w.event(events.VPaintEvent())

    def _exitCleanup(self):
        self._key_event_thread.stop_event.set()
        self._screen.reset()
        super().exit()




from .CommandBarController import CommandBarController
from .EditAreaController import EditAreaController
from .EditorController import EditorController
from .SideRulerController import SideRulerController
from .StatusBarController import StatusBarController



import os
import copy
import json
from .. import paths

class EditorState:
    """
    Class containing recoverable state that wants to survive quit/reopen of the editor.
    The state is dumped to a file in json format.
    """

    # We use a singleton to prevent multiple competing accesses to the file.
    _instance = None

    @classmethod
    def instance(cls):
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def __init__(self):
        if self._instance is not None:
            raise Exception("Only one instance allowed")

        self._state = {}
        try:
            with open(paths.stateFile(), "r") as f:
                self._state = json.loads(f.read())
        except:
            self._state = {}

    def setCursorPosForPath(self, absolute_path, cursor_pos):
        if "buffers" not in self._state:
            self._state["buffers"] = []

        buffers = self._state["buffers"]

        for b in buffers:
            if b.get("absolute_path") == absolute_path:
                b["cursor_pos"] = cursor_pos
        else:
            buffers.append( { "absolute_path" : absolute_path,
                              "cursor_pos" : cursor_pos
                            })

    def cursorPosForPath(self, absolute_path):
        buffers = self._state.get("buffers", [])

        for b in buffers:
            if b.get("absolute_path") == absolute_path:
                return tuple(b.get("cursor_pos"))

        return None

    def save(self):
        """Store the contained state onto the state file"""
        with open(paths.stateFile(), "w") as f:
            f.write(json.dumps(self._state))




from vaitk import core

class EditAreaModel(core.VObject):
    """
    Model defining data pertinent to the Edit Area View.
    """
    def __init__(self):
        super().__init__()
        self._document_pos_at_top = (1,1)
        self.documentPosAtTopChanged = core.VSignal(self)

    @property
    def document_pos_at_top(self):
        """The document position (in document index) in the top-left corner of the editor"""
        return self._document_pos_at_top

    @document_pos_at_top.setter
    def document_pos_at_top(self, doc_pos):
        if doc_pos[0] < 1 or doc_pos[1] < 1:
            raise ValueError("document pos cannot be < 1")

        self._document_pos_at_top = doc_pos
        self.documentPosAtTopChanged.emit()



class VSize(object):
    def __init__(self, width, height):
        self._size = (width, height)

    def __iter__(self):
        return iter(self._size)

    @property
    def width(self):
        return self._size[0]

    @property
    def height(self):
        return self._size[1]

    def __str__(self):
        return "VSize(width=%d, height=%d)" % self._size

    class tuple:
        @staticmethod
        def width(size):
            return size[0]
        
        @staticmethod
        def height(size):
            return size[1]



from yapsy.IPlugin import IPlugin

class CommandPlugin(IPlugin):
    """
    A base class to implement command plugins.

    A command plugin is invoked when the : key is used. The first
    word after the colon is the plugin keyword. When the keyword matches,
    the method execute() is called
    """

    def __init__(self):
        pass

    def name(self):
        """Returns the name of the plugin. To be reimplemented"""
        return None

    def keyword(self):
        """The keyword to activate the plugin."""
        return None

    def execute(self, command_line):
        """
        Method that handles the command.
        It receives the full command line as a string, meaning that it can do pretty much everything.
        If it fails, it's responsibility of the plugin to communicate the error condition.
        Return False if execution failed. True if successful.
        """
        return False




from ... import core
from ..VWidget import VWidget
from ..VPainter import VPainter
from ..VPalette import VPalette

class VLabel(VWidget):
    def __init__(self, label="", parent=None):
        super().__init__(parent)
        self._label = label

    def paintEvent(self, event):
        painter = VPainter(self)
        w, h = self.size()
        painter.fg_color = self.palette().color(VPalette.ColorGroup.Active,
                                        VPalette.ColorRole.WindowText)
        painter.bg_color = self.palette().color(VPalette.ColorGroup.Active,
                                        VPalette.ColorRole.Window)
        string = ' '*w
        for i in range(0, int(h/2)):
            painter.drawText( (0, i), string)
        painter.drawText( (0, int(h/2)), self._label + ' '*(w-len(self._label)))
        for i in range(1+int(h/2), h):
            painter.drawText( (0, i), string)

    def minimumSize(self):
        return (len(self._label), 1)

    def setText(self, text):
        if text != self._label:
            self._label = text
            self.update()



from .BufferCommand import BufferCommand
from .CommandResult import CommandResult

class JoinWithNextLineCommand(BufferCommand):
    def execute(self):
        document = self._document
       
        if self.savedCursorPos() is None:
            self.saveCursorPos()
        
        pos = self.savedCursorPos() 
        
        if pos[0] == document.numLines():
            return CommandResult(success=False, info=None)

        self._cursor.toPos(pos)

        line_meta = document.lineMetaInfo("Change")

        self.saveLineMemento(pos[0], BufferCommand.MEMENTO_REPLACE)
        self.saveLineMemento(pos[0]+1, BufferCommand.MEMENTO_INSERT)

        document.joinWithNextLine(pos[0])
        if line_meta.data(pos[0]) == None:
            line_meta.setData("modified", pos[0])

        return CommandResult(success=True, info=None)




from .VApplication import VApplication

class VCursor:
    @staticmethod
    def setPos(pos):
        VApplication.vApp.screen().setCursorPos(pos)

    @staticmethod
    def pos():
        return VApplication.vApp.screen().cursorPos()





import os
import shlex

from vaitk import gui
from .. import Search
from .. import linting
from .. import paths
from ..lexer import Lexer
from .. import models
from ..models import commands

from yapsy.PluginManager import PluginManager

class EditorController:
    def __init__(self, editor, global_state, buffer_list):
        self._editor = editor
        self._global_state = global_state
        self._buffer_list = buffer_list
        self._lexer = Lexer()
        self._plugin_manager = PluginManager()
        self._plugin_manager.getPluginLocator().setPluginInfoExtension("ini")
        self._plugin_manager.setPluginPlaces([paths.pluginsDir("user", "commands"), paths.pluginsDir("system", "commands")])
        self._plugin_manager.collectPlugins()

        for plugin_info in self._plugin_manager.getAllPlugins():
            self._plugin_manager.activatePluginByName(plugin_info.name)

        # To speed up resolution, we cache the keyword->plugin association. It is filled lazy
        self._keyword_to_plugin_cache = {}

        self._buffer_list.currentBufferChanged.connect(self.registerCurrentBuffer)

    def registerCurrentBuffer(self, *args):
        self._editor.edit_area.buffer = self._buffer_list.current
        self._editor.status_bar_controller.buffer = self._buffer_list.current
        self._editor.side_ruler_controller.buffer = self._buffer_list.current
        self._editor.info_hover_box.buffer = self._buffer_list.current
        self._lexer.setModel(self._buffer_list.current.document)

    def forceQuit(self):
        for b in self._buffer_list.buffers:
            if b.document.filename() is None:
                continue

            models.EditorState.instance().setCursorPosForPath(
                    os.path.abspath(b.document.filename()),
                    b.cursor.pos)

        models.EditorState.instance().save()
        models.Configuration.save()
        gui.VApplication.vApp.exit()

    def doSave(self):
        self._doSave()
        self._doLint()

    def doSaveAs(self, filename):
        self._doSave(filename)
        self._doLint()

    def doInsertFile(self, filename):
        buffer = self._buffer_list.current

        command = commands.InsertFileCommand(buffer, filename)

        result = command.execute()
        if result.success:
            buffer.command_history.push(command)

    def tryQuit(self):
        if any([b.isModified() for b in self._buffer_list.buffers]):
            self._editor.status_bar.setMessage("Document has been modified. " +
                                               "Use :q! to quit without saving " +
                                               "or :qw to save and quit.", 3000)
        else:
            self.forceQuit()

    def searchForward(self, search_text):
        if search_text == '':
            if self._global_state.current_search is not None:
                search_text = self._global_state.current_search[0]

        if search_text != '':
            self._global_state.current_search = (search_text, Search.SearchDirection.FORWARD)
            Search.find(self._buffer_list.current, search_text, Search.SearchDirection.FORWARD)

    def searchBackward(self, search_text):
        if search_text == '':
            if self._global_state.current_search is not None:
                search_text = self._global_state.current_search[0]

        if search_text != '':
            self._global_state.current_search = (search_text, Search.SearchDirection.BACKWARD)
            Search.find(self._buffer_list.current, search_text, Search.SearchDirection.BACKWARD)

    def selectPrevBuffer(self):
        self._buffer_list.selectPrev()

    def selectNextBuffer(self):
        self._buffer_list.selectNext()

    def doSaveAndExit(self):
        self._doSave()
        self.forceQuit()

    def openFile(self, filename):
        buffer = self._buffer_list.bufferForFilename(filename)
        if buffer is not None:
            self._buffer_list.select(buffer)
            return

        current_buffer = self._buffer_list.current
        new_buffer = models.Buffer()
        status_bar = self._editor.status_bar

        try:
            new_buffer.document.open(filename)
        except FileNotFoundError:
            new_buffer.document.setFilename(filename)
            status_bar.setMessage("%s [New file]" % filename, 3000)
        except Exception as e:
            new_buffer.document.setFilename(filename)
            status_bar.setMessage("%s [Error: %s]" % (filename, str(e)), 3000)

        if current_buffer.isEmpty() and not current_buffer.isModified():
            self._buffer_list.replaceAndSelect(current_buffer, new_buffer)
        else:
            self._buffer_list.addAndSelect(new_buffer)

        recovered_cursor_pos = models.EditorState.instance().cursorPosForPath(os.path.abspath(filename))
        if recovered_cursor_pos is not None:
            new_buffer.cursor.toPos(recovered_cursor_pos)

        self._doLint()

    def createEmptyBuffer(self):
        self._buffer_list.addAndSelect(models.Buffer())

    def setMode(self, mode):
        self._global_state.edit_mode = mode

    def interpretCommandLine(self, command_line):
        """
        Interprets and dispatch the command line to the plugin system
        (for now. In the future this controller will just handle all commandline execution)
        command_line contains the full command line as specified by the user, as a string.
        """

        command_tokens = shlex.split(command_line)
        keyword = command_tokens[0]

        if keyword not in self._keyword_to_plugin_cache:
            for plugin_info in self._plugin_manager.getAllPlugins():
                plugin_object = plugin_info.plugin_object
                if plugin_object.keyword() == keyword:
                    self._keyword_to_plugin_cache[keyword] = plugin_object

        plugin_object = self._keyword_to_plugin_cache[keyword]
        plugin_object.execute(command_line)

        # Always return True, regardless. Even if the plugin execution
        # fails, the command was parsed and interpreted correctly
        return True

    # Private

    def _doLint(self):
        document = self._buffer_list.current.document

        linter1 = linting.PyFlakesLinter(document)
        all_info = linter1.runOnce()

        meta_info = {}

        for info in all_info:
            meta_info[info.line] = info

        line_info = document.lineMetaInfo("LinterResult")
        line_info.clear()
        line_info.setDataForLines(meta_info)

    def _doSave(self, filename=None):
        status_bar = self._editor.status_bar
        document = self._buffer_list.current.document

        if filename is not None and len(filename) == 0:
            status_bar.setMessage("Error! Unspecified file name.", 3000)
            return

        status_bar.setMessage("Saving...")
        gui.VApplication.vApp.processEvents()

        try:
            if filename:
                document.saveAs(filename)
            else:
                document.save()
        except models.TextDocument.MissingFilenameException:
            status_bar.setMessage("Error! Cannot save unnamed file. Please specify a filename with :w filename", 3000)
            return
        except Exception as e:
            status_bar.setMessage("Error! Cannot save file. %s" % str(e), 3000)
            return
        else:
            status_bar.setMessage("Saved %s" % document.filename(), 3000)

        document.lineMetaInfo("Change").clear()




from . import VTimerEvent
import logging


class VObject:
    """
    Base class for all objects in VaiTk.
    Provides methods for the object hierarchy traversal.

    A derived class can set the class variable debug to True to enable
    debugging at the DEBUG level for that specific class.

    Example:
    class Foo(VObject):
        debug = True

    The class can also specify the specific logging level
    class Foo(VObject):
        debug = True
        debug_level = logging.INFO

    An instance of VObject will always have a logger in self.logger
    """

    def __new__(cls, *args, **kwargs):
        instance = object.__new__(cls)
        instance.logger = logging.getLogger(cls.__name__)
        if hasattr(cls, "debug"):
            if hasattr(cls, "debug_level"):
                level = cls.debug_level
            else:
                level = logging.DEBUG
            instance.logger.setLevel(level)
            instance.logger.log(level, "Debugging enabled for "+str(cls.__name__))
        else:
            instance.logger.setLevel(logging.CRITICAL+1)

        return instance

    def __init__(self, parent = None):
        self._parent = parent
        self._children = []
        self._event_filters = []
        if self._parent is not None:
            parent.addChild(self)

    def parent(self):
        return self._parent

    def children(self):
        return self._children

    def addChild(self, child):
        self._children.append(child)

    def removeChild(self, child):
        self._children.remove(child)

    def depthFirstFullTree(self):
        return self.root().depthFirstSubTree()

    def depthFirstSubTree(self):
        result = [self]
        for c in self.children():
            result.extend(c.depthFirstSubTree())
        return result

    def root(self):
        return self.traverseToRoot()[-1]

    def traverseToRoot(self):
        result = [self]
        if self.parent() is None:
            return result
        result.extend(self.parent().traverseToRoot())
        return result

    def depthFirstRightTree(self):
        depth_first_tree = self.depthFirstFullTree()
        return depth_first_tree[depth_first_tree.index(self)+1:]

    def depthFirstLeftTree(self):
        depth_first_tree = self.depthFirstFullTree()
        return depth_first_tree[:depth_first_tree.index(self)]

    def installEventFilter(self, event_filter):
        self._event_filters.append(event_filter)

    def eventFilter(self, watched, event):
        return False

    def installedEventFilters(self):
        return self._event_filters

    def event(self, event):
        if isinstance(event, VTimerEvent.VTimerEvent):
            self.timerEvent(event)
            return True
        return False

    def timerEvent(self, event):
        return True

    def __str__(self):
        return self.__class__.__name__



from ... import FocusPolicy
from ... import core
from ...consts import Index
from ... import Key
from ..VWidget import VWidget
from ..VPainter import VPainter
from ..VCursor import VCursor

class VLineEdit(VWidget):
    def __init__(self, contents="", parent=None):
        super().__init__(parent)
        self._text = contents
        self._cursor_position = len(self._text)
        self._selection = None
        self._max_length = 32767
        self.setFocusPolicy(FocusPolicy.StrongFocus)

        self.returnPressed = core.VSignal(self)
        self.cursorPositionChanged = core.VSignal(self)
        self.textChanged = core.VSignal(self)
        self.selectionChanged = core.VSignal(self)
        self.editingFinished = core.VSignal(self)

    def maxLength(self):
        return self._max_length

    def setMaxLength(self, max_length):
        self._max_length = max_length
        self._text = self._text[:self._max_length]
        self.deselect()

    def cursorPosition(self):
        return self._cursor_position

    def setCursorPosition(self, position):
        old_pos = self._cursor_position
        self._cursor_position = position
        self.cursorPositionChanged.emit(old_pos, position)

    def setSelection(self, start, length):
        if len(self._text) == 0:
            return
        self._selection = (0, len(self._text))
        self.selectionChanged.emit()

    def selectAll(self):
        if len(self._text) == 0:
            return
        self._selection = (0, len(self._text))
        self.selectionChanged.emit()

    def selectionStart(self):
        pass

    def selectionEnd(self):
        pass

    def sizeHint(self):
        pass

    def deselect(self):
        self._selection = None
        self.selectionChanged.emit()

    def home(self):
        old_pos = self._cursor_position
        self._cursor_position = 0
        self.cursorPositionChanged.emit(old_pos, self._cursor_position)

    def end(self):
        old_pos = self._cursor_position
        self._cursor_position = len(self._text)
        self.cursorPositionChanged.emit(old_pos, self._cursor_position)

    def text(self):
        return self._text

    def setText(self, text):
        self.deselect()
        if text != self._text:
            self._text = text
            self.textChanged.emit(self._text)
            self.update()

    def backspace(self):
        if self._selection:
            pass
        else:
            pass

    def clear(self):
        self.setText("")
        self._cursor_position = 0

    def cursorForward(self, mark):
        pass

    def cursorBackward(self, mark):
        pass

    def cursorWordForward(self, mark):
        pass

    def cursorWordBackward(self, mark):
        pass

    def minimumSizeHint(self):
        return core.VSize(len(self._text), 1)

    def paintEvent(self, event):
        w, h = self.size()
        painter = VPainter(self)
        painter.drawText( (0, 0), self._text + ' '*(w-len(self._text)))
        if self.hasFocus():
            abs_top_left = self.mapToGlobal((0,0))
            VCursor.setPos( (abs_top_left[Index.X] + self._cursor_position,
                             abs_top_left[Index.Y]
                            )
                          )

    def focusInEvent(self, event):
        abs_top_left = self.mapToGlobal((0,0))
        VCursor.setPos( (abs_top_left[Index.X] + self._cursor_position,
                         abs_top_left[Index.Y]
                        )
                      )

    def keyEvent(self, event):
        if event.key() == Key.Key_Return:
            self.returnPressed.emit()
        elif event.key() == Key.Key_Left:
            self._cursor_position = max(0, self._cursor_position-1)
        elif event.key() == Key.Key_Right:
            self._cursor_position = min(len(self._text), self._cursor_position+1)
        elif event.key() == Key.Key_Backspace:
            if self._cursor_position == 0:
                event.accept()
                return
            self._cursor_position -= 1
            self._text = self._text[:self._cursor_position] + self._text[self._cursor_position+1:]
        else:
            self._text = self._text[:self._cursor_position] + event.text() +  self._text[self._cursor_position:]
            self._cursor_position += len(event.text())
        event.accept()
        self.update()

    def minimumSize(self):
        return (len(self._text), 1)

    def selectedText(self):
        pass



from .BufferCommand import BufferCommand
from .CommandResult import CommandResult
import copy

class DeleteLineAtCursorCommand(BufferCommand):
    def execute(self):
        document = self._document
        if document.isEmpty():
            return CommandResult(success=False, info=None)

        cursor = self._cursor

        if self.savedCursorPos() is None:
            self.saveCursorPos()

        pos = self.savedCursorPos()
        cursor.toPos(pos)

        self.saveLineMemento(pos[0], BufferCommand.MEMENTO_INSERT)
        old_line = copy.deepcopy(self.lastSavedMemento()[2])
        self._old_line_meta_info = {}

        if pos[0] == document.numLines():
            cursor.toLinePrev()
            cursor.toLineBeginning()

        if document.lineMetaInfo("Change").data(pos[0]) != "added":
            # Add markers above and below
            if document.hasLine(pos[0]-1):
                self._old_line_meta_info[-1] = document.lineMetaInfo("Change").data(pos[0]-1)
                document.lineMetaInfo("Change").setData("deletion_before", pos[0]-1)

            if document.hasLine(pos[0]+1):
                self._old_line_meta_info[1] = document.lineMetaInfo("Change").data(pos[0]+1)
                document.lineMetaInfo("Change").setData("deletion_after", pos[0]+1)

        # Delete the line
        document.deleteLine(pos[0])

        # Deleted line, now we check the length of what comes up from below.
        # and set the cursor at the end of the line, if needed
        if document.lineLength(cursor.line) < cursor.column:
            cursor.toPos((cursor.line, document.lineLength(cursor.line)))

        return CommandResult(success=True, info=old_line)

    def undo(self):
        super().undo()
        document = self._document
        cursor = self._cursor
        pos = cursor.pos

        if -1 in self._old_line_meta_info:
            document.lineMetaInfo("Change").setData(self._old_line_meta_info[-1], pos[0]-1)

        if 1 in self._old_line_meta_info:
            document.lineMetaInfo("Change").setData(self._old_line_meta_info[1], pos[0]+1)




class CommandHistory:
    def __init__(self):
        self._past = []
        self._future = []

    def add(self, command):
        self._past.append(command)
        self._future = []

    def prev(self):
        command = self._past.pop(-1)
        self._future.insert(0, command)
        return command

    def next(self):
        command = self._future.pop(0)
        self._past.append(command)
        return command

    def numUndoableCommands(self):
        return len(self._past)

    def numRedoableCommands(self):
        return len(self._future)



from .. import core
from ..consts import Index
from .. import FocusPolicy
from .VApplication import VApplication
from .VPalette import VPalette
from .VPainter import VPainter
from .VScreen import VScreenArea
from . import events


class VWidget(core.VObject):
    def __init__(self, parent=None):
        if parent is None:
            parent = VApplication.vApp.rootWidget()

        super().__init__(parent)

        if self.parent() is None:
            self._geometry = (0,0) + VApplication.vApp.screen().size()
        else:
            self._geometry = self.parent().contentsRect()

        self._layout = None
        self._visible_implicit = False
        self._visible_explicit = None
        self._palette = None
        self._enabled = True
        self._active = True
        self._focus_policy = FocusPolicy.NoFocus
        self._needs_update = False

    def keyEvent(self, event):
        pass

    def setFocus(self):
        VApplication.vApp.setFocusWidget(self)

    def hasFocus(self):
        return (self is VApplication.vApp.focusWidget())

    def move(self, pos):
        if not isinstance(pos, tuple):
            raise TypeError("Invalid pos argument")
        self.setGeometry(pos+self.size())

    def resize(self, size):
        if not isinstance(size, tuple):
            raise TypeError("Invalid size argument")

        self.setGeometry(self.pos()+size)

    def pos(self):
        geometry = self.geometry()
        return ( geometry[Index.RECT_X],
                 geometry[Index.RECT_Y]
                 )


    def size(self):
        return ( self.geometry()[Index.RECT_WIDTH],
                 self.geometry()[Index.RECT_HEIGHT]
                 )

    def rect(self):
        return (0,0)+self.size()

    def absoluteRect(self):
        return self.mapToGlobal((0,0))+self.size()

    def geometry(self):
        return self._geometry

    def width(self):
        return self.geometry()[Index.RECT_WIDTH]

    def height(self):
        return self.geometry()[Index.RECT_HEIGHT]

    def show(self):
        self.setVisible(True)

    def hide(self):
        self.setVisible(False)

    def setVisible(self, visible):
        self.logger.info("Setting explicit visibility for %s : %s" % (str(self), str(visible)))
        visible_before = self.isVisible()
        self._visible_explicit = visible

        if visible and not visible_before:
            VApplication.vApp.postEvent(self,events.VShowEvent())
        elif not visible and visible_before:
            VApplication.vApp.postEvent(self,events.VHideEvent())

        for w in self.children():
            w.setVisibleImplicit(visible)

    def setVisibleImplicit(self, visible):
        self.logger.info("Setting implicit visibility for %s : %s" % (str(self), str(visible)))
        self._visible_implicit = visible

        if visible:
            VApplication.vApp.postEvent(self,events.VShowEvent())
        else:
            VApplication.vApp.postEvent(self,events.VHideEvent())

        for w in self.children():
            w.setVisibleImplicit(visible)

    def isVisible(self):
        return self._visible_explicit if self._visible_explicit is not None else self._visible_implicit

    def minimumSize(self):
        return (0,0)

    def addLayout(self, layout):
        self._layout = layout
        self._layout.setParent(self)

    def setGeometry(self, rect):
        self.logger.info("VWidget.setGeometry %s" % str(rect))

        old_geometry = self._geometry

        min_size = self.minimumSize()
        self._geometry = (rect[Index.RECT_X],
                          rect[Index.RECT_Y],
                          max(min_size[Index.SIZE_WIDTH], rect[Index.RECT_WIDTH]),
                          max(min_size[Index.SIZE_HEIGHT], rect[Index.RECT_HEIGHT])
                         )

        if not self.isVisible():
            return

        if  (old_geometry[Index.RECT_X], old_geometry[Index.RECT_Y])  \
            != (self._geometry[Index.RECT_X], self._geometry[Index.RECT_Y]):

            VApplication.vApp.postEvent(self, events.VMoveEvent())

        if (old_geometry[Index.RECT_WIDTH], old_geometry[Index.RECT_WIDTH]) \
            != (self._geometry[Index.RECT_HEIGHT], self._geometry[Index.RECT_HEIGHT]):

            VApplication.vApp.postEvent(self, events.VResizeEvent())

    def mapToGlobal(self, pos):
        top_left = self.pos()
        if self.parent() is None:
            return (pos[Index.X]+top_left[Index.X], pos[Index.Y]+top_left[Index.Y])

        parent_corner = self.parent().mapToGlobal((0,0))
        return ( parent_corner[Index.X] + top_left[Index.X] + pos[Index.X],
                 parent_corner[Index.Y] + top_left[Index.Y] + pos[Index.Y]
                 )

    def screenArea(self):
        abs_pos_topleft = self.mapToGlobal((0,0))

        return VScreenArea( VApplication.vApp.screen(),
                            abs_pos_topleft + self.size()
                          )

    def event(self, event):
        self.logger.info("Event %s. Receiver %s" % (str(event), str(self)))

        if isinstance(event, events.VPaintEvent):
            if not self.isVisible():
                return True
            self.paintEvent(event)
            self._needs_update = False

        elif isinstance(event, events.VFocusEvent):
            if self.isVisible():
                if event.eventType() == core.VEvent.EventType.FocusIn:
                    self.focusInEvent(event)
            else:
                if event.eventType() == core.VEvent.EventType.FocusOut:
                    self.focusOutEvent(event)

            self.update()

        elif isinstance(event, events.VHideEvent):
            self.hideEvent(event)

            for w in self.depthFirstFullTree():
                self.logger.info("Widget %s in tree" % str(w))
                if not w.isVisible():
                    continue
                self.logger.info("Repainting widget %s" % str(w))
                w.update()

        elif isinstance(event, events.VShowEvent):
            self.showEvent(event)

            for w in self.depthFirstFullTree():
                self.logger.info("Widget %s in tree" % str(w))
                if not w.isVisible():
                    continue
                w.update()

        elif isinstance(event, events.VMoveEvent):
            if not self.isVisible():
                return True

            self.moveEvent(event)

            for w in self.depthFirstFullTree():
                self.logger.info("Widget %s in tree" % str(w))
                if not w.isVisible():
                    continue
                w.update()

        elif isinstance(event, events.VResizeEvent):
            if not self.isVisible():
                return True

            self.resizeEvent(event)

            for w in self.depthFirstFullTree():
                self.logger.info("Widget %s in tree" % str(w))
                if not w.isVisible():
                    continue
                w.update()
        else:
            return super().event(event)

        return True

    def paintEvent(self, event):
        painter = VPainter(self)
        #if self._layout is not None:
        #    self._layout.apply()

        size = self.size()

        string = ' '*size[Index.SIZE_WIDTH]
        for i in range(0, size[Index.SIZE_HEIGHT]):
            painter.drawText( (0, i), string)
    def needsUpdate(self):
        return self._needs_update

    def focusInEvent(self, event):
        self.logger.info("FocusIn event")

    def focusOutEvent(self, event):
        self.logger.info("FocusOut event")

    def hideEvent(self, event):
        self.logger.info("Hide event")

    def moveEvent(self, event):
        self.logger.info("Move event")

    def showEvent(self, event):
        self.logger.info("Show event")

    def resizeEvent(self, event):
        self.logger.info("Resize event")

    def setFocusPolicy(self, policy):
        self._focus_policy = policy

    def focusPolicy(self):
        return self._focus_policy

    def isEnabled(self):
        return self._enabled

    def isActive(self):
        return self._active

    def setActive(self, active):
        self._active = active

    def setEnabled(self, enabled):
        self._enabled = enabled

    def palette(self):
        if self._palette is None:
            self._palette = VApplication.vApp.palette().copy()

        return self._palette

    def setColors(self, fg=None, bg=None):
        self.palette().setColor(VPalette.ColorGroup.Active, VPalette.ColorRole.WindowText, fg)
        self.palette().setColor(VPalette.ColorGroup.Active, VPalette.ColorRole.Window, bg)

    def colors(self, color_group = VPalette.ColorGroup.Active):
        fg = self.palette().color(color_group, VPalette.ColorRole.WindowText)
        bg = self.palette().color(color_group, VPalette.ColorRole.Window)

        return (fg, bg)

    def currentColors(self):
        if self.isActive():
            color_group = VPalette.ColorGroup.Active
        else:
            if isEnabled(self):
                color_group = VPalette.ColorGroup.Inactive
            else:
                color_group = VPalette.ColorGroup.Disabled
        return self.colors(color_group)

    def update(self):
        self._needs_update = True

    def contentsRect(self):
        margins = self.contentsMargins()
        return (margins[Index.MARGIN_LEFT],
                margins[Index.MARGIN_TOP],
                self.width()-margins[Index.MARGIN_LEFT]-margins[Index.MARGIN_RIGHT],
                self.height()-margins[Index.MARGIN_TOP]-margins[Index.MARGIN_BOTTOM]
                )

    def contentsMargins(self):
        return (0,0,0,0)



from vaitk import gui
from yapsy.PluginManager import PluginManager
import collections
from .. import paths
from ..lexer import token as lexertoken

class SyntaxColors:
    def __init__(self, schema_name, num_colors):
        self._color_map = collections.defaultdict(lambda : (None, None, None))

        if schema_name != "default" and self._tryLoad(schema_name, num_colors):
            return

        self._installDefault(num_colors)

    def _tryLoad(self, schema_name, num_colors):
        plugin_manager = PluginManager()
        plugin_manager.getPluginLocator().setPluginInfoExtension("ini")
        plugin_manager.setPluginPlaces([paths.pluginsDir("user", "syntaxcolors"), paths.pluginsDir("system", "syntaxcolors")])
        plugin_manager.collectPlugins()

        for plugin_info in plugin_manager.getAllPlugins():

            # Useless, but let's activate them
            plugin_manager.activatePluginByName(plugin_info.name)

            plugin_object = plugin_info.plugin_object
            if plugin_object.name() == schema_name and plugin_object.supportsNumColors(num_colors):

                self._color_map.update(_parseColorSchema(plugin_object.colorSchema(num_colors)))
                return True

        return False

    def _installDefault(self, num_colors):
        self._color_map.update(_parseColorSchema(defaultColorSchema(num_colors)))

    def colorMap(self):
        return self._color_map

def defaultColorSchema(num_colors):
    if num_colors == 8:
        return {
            "Keyword":              "yellow",
            "Keyword.Constant":     "red",
            "Keyword.Pseudo":       "red",
            "Keyword.Namespace":    "magenta",
            "Keyword.Reserved":     "magenta",
            "Keyword.Type":         "magenta",
            "Comment":              "cyan",
            "Comment.Single":       "cyan",
            "Name.Function":        "cyan",
            "Name.Class":           "cyan",
            "Name.Builtin":         "cyan",
            "Name.Exception":       "cyan",
            "String":               "red",
            "Literal":              "red",
            "Literal.String.Doc":   "red",
            "Number":               "red",
            "Number.Integer":       "red",
            "Number.Float":         "red",
            "Number.Hex":           "red",
            "Number.Oct":           "red",
            "Operator.Word":        "yellow",
            "Name.Decorator":       "blue",
            "Name.Builtin.Pseudo":  "green",
        }

    elif num_colors == 256:
        return {
            "Keyword":                        "term_ffd75f",
            "Keyword.Constant":               "lightred",
            "Keyword.Pseudo":                 "lightred",
            "Keyword.Namespace":              "term_af87af",
            "Keyword.Reserved":               "term_af87af",
            "Keyword.Type":                   "term_af87af",
            "Comment":                        "cyan",
            "Comment.Single":                 "cyan",
            "Name.Class":                     "lightcyan",
            "Name.Class.PythonPrivate":       "term_ff5f5f",
            "Name.Builtin":                   "lightcyan",
            "Name.Exception":                 "lightcyan",
            "String":                         "lightred",
            "Literal":                        "lightred",
            "Literal.String.Doc":             "lightred",
            "Number":                         "lightred",
            "Number.Integer":                 "lightred",
            "Number.Float":                   "lightred",
            "Number.Hex":                     "lightred",
            "Number.Oct":                     "lightred",
            "Operator.Word":                  "term_ffd75f",
            "Name.Decorator":                 "blue",
            "Name.Builtin.Pseudo":            "term_5fafd7",
            "Name.Builtin.Pseudo.PythonSelf": "pink",
            "Name.Function":                  "term_5fff5f",
            "Name.Function.PythonPrivate":    "term_ff5f5f",
            "Name.Function.PythonMagic":      "term_5f5fff",
        }

def _parseColorSchema(color_schema_dict):
    result = {}
    for tok, col in color_schema_dict.items():
        result[_token(tok)] = _color(col)

    return result

def _token(token_string):
    """
    Returns a representation useful for the SyntaxColors Plugin.
    token_string is a string describing the token. The names are the one provides
    by pygments, with the addition of some vai specific tokens

    To return a representation of the Keyword token

        token("Keyword")

    To return a representation of the Keyword.Constant token

        token("Keyword.Constant")

    The returned representation is an implementation detail.
    """
    subtokens = token_string.split(".")

    if len(subtokens) == 0:
        return None

    target = lexertoken
    for subtoken in subtokens:
        new_target = target.__dict__.get(subtoken)
        if new_target is None:
            return None
        target = new_target

    return target

def _color(color_string):
    """
    Ease method to produce a representation of a color for the syntax highlighting.
    It can be called in the following ways

    To return a representation of the color blue for the foreground

        color("blue")

    To return a representation of the color blue, with bold character

        color("blue bold")

    To return a representation of blue foreground with white background

        color("blue,white")

    To return a representation of blue bold foreground with white background

        color("blue bold,white")
    """

    fg_bg = color_string.split(",")

    fg, bg_color = fg_bg if len(fg_bg) == 2 else (fg_bg[0], None)
    if bg_color is not None:
        bg_color = bg_color.strip()

    fg_font = fg.split(" ")

    fg_color, fg_font = fg_font if len(fg_font) == 2 else (fg_font[0], None)

    fg_color = fg_color.strip()
    if fg_font is not None:
        fg_font = fg_font.strip()

    return (gui.VGlobalColor.__dict__.get(fg_color), None, gui.VGlobalColor.__dict__.get(bg_color))




from .. import models
from vaitk import gui, core, utils

class StatusBar(gui.VLabel):
    def __init__(self, parent):
        super().__init__("", parent)
        self._filename = ""
        self._file_changed_flag = False
        self._position = ""
        self._message = ""
        self._remove_message_timer = core.VTimer()
        self._remove_message_timer.setSingleShot(True)
        self._remove_message_timer.timeout.connect(self.clearMessage)


        fg_color = gui.VGlobalColor.nameToColor(models.Configuration.get("colors.status_bar.fg"))
        bg_color = gui.VGlobalColor.nameToColor(models.Configuration.get("colors.status_bar.bg"))
        
        self.setColors(fg_color, bg_color)

    def setMessage(self, message, timeout=None):
        if self._remove_message_timer.isRunning():
            self._remove_message_timer.stop()

        if message is None:
            self.clearMessage()
            return

        self._message = message
        self._updateText()

        if timeout is not None:
            self._remove_message_timer.setInterval(timeout)
            self._remove_message_timer.start()

    def clearMessage(self):
        self._message = ""
        self._updateText()

    def setFilename(self, filename):
        self._filename = filename
        self._updateText()

    def setPosition(self, pos):
        self._position = str(pos[0])+","+str(pos[1])
        self._updateText()

    def setFileChangedFlag(self, changed):
        self._file_changed_flag = changed
        self._updateText()

    # Private
    def _updateText(self):
        if self._filename == None:
            filename = "[No Name]"
        else:
            filename = self._filename

        if len(filename) > int(self.width()/2):
            filename = "<"+filename[-int(self.width()/2):]

        if self._message == "":
            self.setText(utils.strformat([(0, filename),
                                          (len(filename)+1, '[+]' if self._file_changed_flag else '   '),
                                          (self.width()-len(self._position)-3, self._position)
                                         ], self.width()))
        else:
            self.setText(utils.strformat([(0, self._message)], self.width()))




from .BufferCommand import BufferCommand
from .CommandResult import CommandResult

class DeleteToEndOfWordCommand(BufferCommand):
    SPACERS = " {}[]().!@#$%^&*=,"

    def execute(self):
        cursor = self._buffer.cursor
        document = self._buffer.document

        if self.savedCursorPos() is None:
            self.saveCursorPos()

        pos = self.savedCursorPos()
        cursor.toPos(pos)

        self.saveLineMemento(pos[0], BufferCommand.MEMENTO_REPLACE)

        line_meta = document.lineMetaInfo("Change")
        changed = line_meta.data(pos[0])
        if changed is None:
            line_meta.setData("modified", pos[0])

        text = document.lineText(pos[0])

        # Zero based, for text.
        cur_index = pos[1]-1

        char_under_cursor = text[cur_index]
        if text[cur_index] in DeleteToEndOfWordCommand.SPACERS:
            # cursor is on a spacer. Delete everything up to the first non spacer
            # or the first spacer that is not the one we are currently on.
            remove_count = len(text[cur_index:]) - len(text[cur_index:].lstrip(char_under_cursor))
            remove_count += len(text[cur_index+remove_count:]) - len(text[cur_index+remove_count:].lstrip(' '))
        else:
            # Get the next spacer from the current position.
            # It's the first of this list comprehension, if available
            spacer_indexes = [ p for p, c in enumerate(text) if (p > cur_index and c in DeleteToEndOfWordCommand.SPACERS)]
            if len(spacer_indexes) == 0:
                # None, just remove to end of line
                remove_count = document.lineLength(pos[0]) - cur_index + 1
            else:
                # Remove the non-spacer, plus all the ' ' up to the next spacer
                remove_count = spacer_indexes[0] - cur_index
                remove_count += len(text[cur_index+remove_count:]) - len(text[cur_index+remove_count:].lstrip(' '))

        deleted = document.deleteChars(pos, remove_count)
        return CommandResult(success=True, info=deleted)






def strformat(strings_at, length):
    """
    Returns a string of given length, composed by strings at specified
    positions, as specified by strings_at. The rest is filled with spaces.
    strings_at is a list of tuples of two elements, the first a number,
    the second a string.
    """

    res = ""
    for pos, string in sorted(strings_at, key=lambda x: x[0]):
        res = res[:pos]
        res += " "* (pos - len(res))
        res += string

    res = res[:length]
    res += " "* (length - len(res))
    return res

def clamp(value, minvalue, maxvalue):
    """
    Returns the value, clamped between min and max value
    if outside that range
    """
    return sorted([minvalue, value, maxvalue])[1]



import shlex
from .. import models
import sys
import os

class CommandBarController:
    def __init__(self, command_bar, edit_area, editor_controller, global_state):
        self._command_bar = command_bar
        self._edit_area = edit_area
        self._editor_controller = editor_controller
        self._global_state = global_state

        self._command_bar.returnPressed.connect(self.parseCommandBar)
        self._command_bar.escapePressed.connect(self.abortCommandBar)
        self._command_bar.tabPressed.connect(self.autocompleteCommandBar)

        self._global_state.editorModeChanged.connect(self.editorModeChanged)

    def parseCommandBar(self):
        command_text = self._command_bar.command_text
        mode = self._global_state.editor_mode
        self._global_state.editor_mode = models.EditorMode.COMMAND

        if mode == models.EditorMode.COMMAND_INPUT:
            if self._interpretLine(command_text):
                self._command_bar.clear()
        elif mode == models.EditorMode.SEARCH_FORWARD:
            self._editor_controller.searchForward(command_text)
            self._command_bar.clear()
        elif mode == models.EditorMode.SEARCH_BACKWARD:
            self._editor_controller.searchBackward(command_text)
            self._command_bar.clear()

        self._edit_area.setFocus()

    def abortCommandBar(self):
        self._command_bar.clear()
        self._global_state.editor_mode = models.EditorMode.COMMAND
        self._edit_area.setFocus()

    def editorModeChanged(self, *args):
        self._command_bar.editor_mode = self._global_state.editor_mode

    def _interpretLine(self, command_text):

        # FIXME all this code must be moved to the editor controller.
        if len(command_text.strip()) == 0:
            return True

        command = shlex.split(command_text)
        if command[0] == 'q!':
            self._editor_controller.forceQuit()
        elif command[0] == 'q':
            self._editor_controller.tryQuit()
        elif command[0] == "w":
            if len(command) == 1:
                self._editor_controller.doSave()
            elif len(command) == 2:
                self._editor_controller.doSaveAs(command[1])
            else:
                self._reportError("Only one filename allowed at write")
                return False
        elif command[0] == "r":
            if len(command) == 1:
                self._reportError("Specify filename")
                return False
            elif len(command) == 2:
                self._editor_controller.doInsertFile(command[1])
            else:
                self._reportError("Only one filename allowed")
                return False
        elif command[0] in ("wq", "x"):
            self._editor_controller.doSaveAndExit()
        elif command[0] == "e":
            if len(command) == 1:
                self._reportError("Specify filename")
            elif len(command) == 2:
                self._editor_controller.openFile(command[1])
            else:
                self._reportError("Only one filename allowed")
                return False
        elif command[0] == "bp":
            self._editor_controller.selectPrevBuffer()
        elif command[0] == "bn":
            self._editor_controller.selectNextBuffer()
        else:
            return self._editor_controller.interpretCommandLine(command_text)
        return True

    def _reportError(self, error_string):
        self._command_bar.setErrorString(error_string)

    def autocompleteCommandBar(self):
        command_text = self._command_bar.command_text
        if len(command_text.strip()) == 0:
            return True

        command = shlex.split(command_text)

        if command[0] in ("w", "r", "e"):
            if len(command) == 2:
                path = command[1]

                dirname = os.path.join(".", os.path.dirname(path))
                basename = os.path.basename(path)
                if os.path.isdir(dirname):
                    files = [f for f in os.listdir(dirname) if f.startswith(basename)]
                    prefix = os.path.commonprefix(files)
                    add_to_basename = ''
                    if len(prefix) > len(basename):
                        add_to_basename = prefix[len(basename):]

                    new_command_text = self._command_bar.command_text+add_to_basename
                    new_path = os.path.join(dirname, basename+add_to_basename)
                    if os.path.isdir(new_path) and new_path[-1] != '/':
                        new_command_text += '/'

                    self._command_bar.command_text = new_command_text




from ..VWidget import VWidget

class VPushButton(VWidget):
    def __init__(self, label, parent=None):
        super(VPushButton, self).__init__(parent)
        self._label = label

    def render(self, painter):
        super(VPushButton, self).render(painter)
        for i in range(0, h/2):
            painter.write(0, i, ' '*w)
        painter.write(0, h/2, "[ "+self._label + " ]"+ ' '*(w-len(self._label)-4))
        for i in range(1+h/2, h):
            painter.write(0, i, ' '*w)



class Index:
    FG_COLOR = 0
    BG_COLOR = 1
    X = 0
    Y = 1
    RECT_X = 0
    RECT_Y = 1
    RECT_WIDTH = 2
    RECT_HEIGHT = 3
    SIZE_WIDTH = 0
    SIZE_HEIGHT = 1
    MARGIN_LEFT = 0
    MARGIN_TOP = 1
    MARGIN_RIGHT = 2
    MARGIN_BOTTOM = 3



from .VApplication import VApplication
from .VColor import VColor, VGlobalColor
from .VCursor import VCursor
from .VPalette import VPalette
from .VPainter import VPainter
from .VScreen import VScreen, VScreenArea
from .VWidget import VWidget
from .widgets import VLabel, VDialog, VLineEdit, VToolTip
from .layouts import VHLayout, VVLayout
from .events import VKeyEvent, VPaintEvent








import itertools

class VTextScreen(object):
    """
    Dummy Screen that renders information in an indexed buffer, instead of the actual terminal
    screen.
    """
    def __init__(self, size):
        self._cursor_pos = (0,0)
        self._size = size
        self._text = ""

        self._render_output = []
        for h in range(self._size[1]):
            row = []
            self._render_output.append(row)
            for w in range(self._size[0]):
                row.append('.')

        self.erase()
        self._log = []
        self._log.append("Inited screen")

    def erase(self):
        self._render_output = []
        for h in range(self._size[1]):
            row = []
            self._render_output.append(row)
            for w in range(self._size[0]):
                row.append('.')

    def deinit(self):
        self._log.append("Deinited screen")

    def refresh(self):
        self._log.append("Refresh")

    def size(self):
        return self._size

    def cursorPos(self):
        return self._cursor_pos

    def setCursorPos(self, pos):
        self._cursor_pos = pos

    def addstr(self, *args):
        pass

    def getch(self, *args):
        if len(self._text) == 0:
            return -1
        char_ret, self._text = self._text[0], self._text[1:]

        return char_ret

    def typeText(self, text):
        self._text = text

    def getColor(self, fg, bg):
        return 0

    def write(self, pos, string, fg_color=None, bg_color=None):
        for pos_x in range(len(string)):
            try:
                self._render_output[pos[1]][pos[0]+pos_x] = string[pos_x]
            except:
                print("Invalid write position : ", str(pos), string)
    def dump(self):
        ret = []
        ret.append(" "+"".join(list(itertools.islice(itertools.cycle(list(map(str, list(range(10))))), self._size[0]+1))))
        #print "+"*(self._size[0]+2)
        for i, r in enumerate(self._render_output):
            ret.append(str(i%10)+''.join(r)+"+")
        ret.append( "+"*(self._size[0]+2))
        return ret

    def __str__(self):
        return "\n".join(self.dump())

    def charAt(self, x, y):
        return self._render_output[y][x]

    def stringAt(self, x, y, l):
        return ''.join(self._render_output[y][x:x+l])

    def numColors(self):
        return 8



#!/usr/bin/env python3.4
import sys

if '3.4' not in sys.version.split()[0]:
    print("Needs python 3.4")
    sys.exit(0)

import os
import codecs
import re
from setuptools import setup, find_packages



base_dir = os.path.abspath(os.path.dirname(__file__))
def find_version(*file_paths):
    # Open in Latin-1 so that we avoid encoding errors.
    # Use codecs.open for Python 2 compatibility
    with codecs.open(os.path.join(base_dir, *file_paths), 'r', 'latin1') as f:
        version_file = f.read()

    # The version line must have the form
    # __version__ = 'ver'
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")

# Get the long description from the relevant file
with codecs.open('DESCRIPTION.rst', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name = "vai",
    version = find_version("vai", "__init__.py"),
    author = "Stefano Borini",
    author_email = "stefano.borini@gmail.com",
    description = "VAI is a console-based IDE similar to VIM",
    license = "BSD",
    keywords = "editor",
    url="https://github.com/stefanoborini/vai",
    packages=find_packages(exclude=["contrib", "docs", "tests*"]),
    long_description=long_description,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Text Editors",
        "License :: OSI Approved :: BSD License",
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3.4',
    ],
    test_suite = 'nose.collector',
    install_requires = ['pygments==2.0.1', "pylint==1.3.0", 'pyflakes==0.8.1', 'yapsy==1.10.423'],
    entry_points={
        'console_scripts': [
            'vai_exec=vai:main',
        ],
    },
    scripts=['bin/vai'],
    zip_safe=False,
)




"""
Module that wraps and augments the pygment tokens, so that we can provide additional
tokens if we recognize them as having additional properties worth of interest.
"""

from pygments.token import *

# Special tokens for python stuff
PythonSelf = Name.Builtin.Pseudo.PythonSelf
PythonPrivate = Name.Function.PythonPrivate
PythonMagic = Name.Function.PythonMagic





from .TextDocument import TextDocument
from .TextDocumentCursor import TextDocumentCursor
from .EditAreaModel import EditAreaModel
from .CommandHistory import CommandHistory

class Buffer:
    """
    Represents an editable buffer, and contains the document, the cursor
    position, the command history, and all the state that is local to a
    specific buffer.
    """
    def __init__(self):
        self._document = TextDocument()
        self._document_cursor = TextDocumentCursor(self._document)
        self._edit_area_model = EditAreaModel()
        self._command_history = CommandHistory()
        self._document.createLineMetaInfo("LinterResult")
        self._document.createLineMetaInfo("Change")
        self._document.createLineMetaInfo("Bookmark")

    def isEmpty(self):
        """
        Returns True if the document is empty
        """
        return self._document.isEmpty()

    def isModified(self):
        """
        Returns True if the document is modified
        """
        return self._document.isModified()

    @property
    def document(self):
        return self._document

    @property
    def cursor(self):
        return self._document_cursor

    @property
    def edit_area_model(self):
        return self._edit_area_model

    @property
    def command_history(self):
        return self._command_history




import vaitk
from vaitk import gui, core
from ..models import EditorMode

class CommandBar(gui.VWidget):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.returnPressed = core.VSignal(self)
        self.escapePressed = core.VSignal(self)
        self.tabPressed = core.VSignal(self)

        self._editor_mode = EditorMode.COMMAND

        self._state_label = gui.VLabel(parent=self)
        self._state_label.setGeometry((0,0,1,1))

        self._line_edit = gui.VLineEdit(parent=self)
        self._line_edit.returnPressed.connect(self.returnPressed)
        self._line_edit.setGeometry((1,0,self.width()-1,1))
        self._line_edit.installEventFilter(self)
        self._updateText()

    @property
    def editor_mode(self):
        return self._editor_mode

    @editor_mode.setter
    def editor_mode(self, editor_mode):
        self._editor_mode = editor_mode
        self._updateText()

    def setEditorMode(self, editor_mode):
        self.editor_mode = editor_mode

    @property
    def command_text(self):
        return self._line_edit.text().strip()

    @command_text.setter
    def command_text(self, text):
        self._line_edit.setText(text)
        self._line_edit.end()


    def clear(self):
        self._editor_mode = EditorMode.COMMAND
        self._line_edit.clear()
        self._updateText()

    def setFocus(self):
        self._line_edit.setFocus()

    def eventFilter(self, event):
        if isinstance(event, gui.VKeyEvent):
            if event.key() == vaitk.Key.Key_Escape:
                self.escapePressed.emit()
                return True
            elif event.key() == vaitk.Key.Key_Backspace and len(self.command_text) == 0:
                self.escapePressed.emit()
                return True
            if event.key() == vaitk.Key.Key_Tab:
                self.tabPressed.emit()
                return True
        return False

    def setErrorString(self, error_string):
        self._line_edit.clear()
        self._state_label.resize( (len(error_string), 1) )
        self._state_label.setText(error_string)
        self._line_edit.setGeometry( (len(error_string), 0, self.width()-len(error_string), 1) )

    # Private

    def _updateText(self):
        if self._editor_mode == EditorMode.INSERT:
            text = "-- INSERT --"
        elif self._editor_mode == EditorMode.COMMAND_INPUT:
            text = "Command: "
        elif self._editor_mode == EditorMode.REPLACE:
            text = "-- REPLACE --"
        elif self._editor_mode == EditorMode.VISUAL_BLOCK:
            text = "-- VISUAL BLOCK --"
        elif self._editor_mode == EditorMode.VISUAL_LINE:
            text = "-- VISUAL LINE --"
        elif self._editor_mode == EditorMode.VISUAL:
            text = "-- VISUAL --"
        elif self._editor_mode == EditorMode.SEARCH_FORWARD:
            text = "Search: "
        elif self._editor_mode == EditorMode.SEARCH_BACKWARD:
            text = "Search backward: "
        elif self._editor_mode == EditorMode.BOOKMARK:
            text = "Set bookmark ..."
        elif self._editor_mode == EditorMode.GOTOBOOKMARK:
            text = "Go to bookmark ..."
        elif self._editor_mode == EditorMode.DELETE:
            text = "Delete ..."
        elif self._editor_mode == EditorMode.GO:
            text = "Go to ..."
        else:
            text = ""

        self._state_label.resize( (len(text), 1) )
        self._state_label.setText(text)
        self._line_edit.setGeometry( (len(text), 0, self.width()-len(text), 1) )




from .BufferCommand import BufferCommand
from .CommandResult import CommandResult

class DeleteSingleCharAfterCommand(BufferCommand):
    def execute(self):
        document = self._document

        if self.savedCursorPos() is None:
            self.saveCursorPos()

        pos = self.savedCursorPos()
        
        self._cursor.toPos(pos)
 
        if pos[1] == document.lineLength(pos[0]):
            return CommandResult(success=False, info=None)

        self.saveLineMemento(pos[0], BufferCommand.MEMENTO_REPLACE)

        line_meta = document.lineMetaInfo("Change")
        changed = line_meta.data(pos[0])

        if changed is None:
            line_meta.setData("modified", pos[0])

        deleted = document.deleteChars(pos, 1)
        return CommandResult(success=True, info=deleted)




from . import VColor
import copy

class VPalette:
    class ColorGroup:
        Active, Disabled, Inactive = list(range(3))

    class ColorRole:
        WindowText, \
        Button, \
        Light, \
        Midlight, \
        Dark, \
        Mid, \
        Text, \
        BrightText, \
        ButtonText, \
        Base, \
        Window, \
        Shadow, \
        Highlight, \
        HighlightedText, \
        Link, \
        LinkVisited, \
        AlternateBase, \
        NoRole, \
        ToolTipBase, \
        ToolTipText = list(range(20))

    def __init__(self):
        self._colors = {}

    def color(self, color_group, color_role):
        return self._colors[(color_group, color_role)]

    def setColor(self, color_group, color_role, color):
        self._colors[(color_group, color_role)] = color

    def setDefaults(self):
        self._colors = {
            ( VPalette.ColorGroup.Active, VPalette.ColorRole.WindowText) : VColor.VColor( rgb = (170,170,170)),
            ( VPalette.ColorGroup.Active, VPalette.ColorRole.Button) : VColor.VColor( rgb = (255,255,255)),
            ( VPalette.ColorGroup.Active, VPalette.ColorRole.Light) : VColor.VColor( rgb = (255,255,255)),
            ( VPalette.ColorGroup.Active, VPalette.ColorRole.Midlight) : VColor.VColor( rgb = (255,255,255)),
            ( VPalette.ColorGroup.Active, VPalette.ColorRole.Dark) : VColor.VColor( rgb = (255,255,255)),
            ( VPalette.ColorGroup.Active, VPalette.ColorRole.Mid) : VColor.VColor( rgb = (255,255,255)),
            ( VPalette.ColorGroup.Active, VPalette.ColorRole.Text) : VColor.VColor( rgb = (170,170,170)),
            ( VPalette.ColorGroup.Active, VPalette.ColorRole.BrightText) : VColor.VColor( rgb = (255,255,255)),
            ( VPalette.ColorGroup.Active, VPalette.ColorRole.ButtonText) : VColor.VColor( rgb = (255,255,255)),
            ( VPalette.ColorGroup.Active, VPalette.ColorRole.Base) : VColor.VColor( rgb = (0,0,0)),
            ( VPalette.ColorGroup.Active, VPalette.ColorRole.Window) : None,
            ( VPalette.ColorGroup.Active, VPalette.ColorRole.Shadow) : VColor.VColor( rgb = (0,0,0)),
            ( VPalette.ColorGroup.Active, VPalette.ColorRole.Highlight) : VColor.VColor( rgb = (255,255,255)),
            ( VPalette.ColorGroup.Active, VPalette.ColorRole.HighlightedText) : VColor.VColor( rgb = (0,0,0)),
            ( VPalette.ColorGroup.Active, VPalette.ColorRole.Link) : VColor.VColor( rgb = (255,255,255)),
            ( VPalette.ColorGroup.Active, VPalette.ColorRole.LinkVisited) : VColor.VColor( rgb = (255,255,255)),
            ( VPalette.ColorGroup.Active, VPalette.ColorRole.AlternateBase) : VColor.VColor( rgb = (255,255,255)),
            ( VPalette.ColorGroup.Active, VPalette.ColorRole.NoRole) : VColor.VColor( rgb = (255,255,255)),
            ( VPalette.ColorGroup.Active, VPalette.ColorRole.ToolTipBase) : VColor.VColor( rgb = (170,0,170)),
            ( VPalette.ColorGroup.Active, VPalette.ColorRole.ToolTipText) : VColor.VColor( rgb = (255,255,0)),
            ( VPalette.ColorGroup.Disabled, VPalette.ColorRole.WindowText) : VColor.VColor( rgb = (255,255,255)),
            ( VPalette.ColorGroup.Disabled, VPalette.ColorRole.Button) : VColor.VColor( rgb = (255,255,255)),
            ( VPalette.ColorGroup.Disabled, VPalette.ColorRole.Light) : VColor.VColor( rgb = (255,255,255)),
            ( VPalette.ColorGroup.Disabled, VPalette.ColorRole.Midlight) : VColor.VColor( rgb = (255,255,255)),
            ( VPalette.ColorGroup.Disabled, VPalette.ColorRole.Dark) : VColor.VColor( rgb = (255,255,255)),
            ( VPalette.ColorGroup.Disabled, VPalette.ColorRole.Mid) : VColor.VColor( rgb = (255,255,255)),
            ( VPalette.ColorGroup.Disabled, VPalette.ColorRole.Text) : VColor.VColor( rgb = (255,255,255)),
            ( VPalette.ColorGroup.Disabled, VPalette.ColorRole.BrightText) : VColor.VColor( rgb = (255,255,255)),
            ( VPalette.ColorGroup.Disabled, VPalette.ColorRole.ButtonText) : VColor.VColor( rgb = (255,255,255)),
            ( VPalette.ColorGroup.Disabled, VPalette.ColorRole.Base) : VColor.VColor( rgb = (0,0,0)),
            ( VPalette.ColorGroup.Disabled, VPalette.ColorRole.Window) : None,
            ( VPalette.ColorGroup.Disabled, VPalette.ColorRole.Shadow) : VColor.VColor( rgb = (0,0,0)),
            ( VPalette.ColorGroup.Disabled, VPalette.ColorRole.Highlight) : VColor.VColor( rgb = (255,255,255)),
            ( VPalette.ColorGroup.Disabled, VPalette.ColorRole.HighlightedText) : VColor.VColor( rgb = (0,0,0)),
            ( VPalette.ColorGroup.Disabled, VPalette.ColorRole.Link) : VColor.VColor( rgb = (255,255,255)),
            ( VPalette.ColorGroup.Disabled, VPalette.ColorRole.LinkVisited) : VColor.VColor( rgb = (255,255,255)),
            ( VPalette.ColorGroup.Disabled, VPalette.ColorRole.AlternateBase) : VColor.VColor( rgb = (255,255,255)),
            ( VPalette.ColorGroup.Disabled, VPalette.ColorRole.NoRole) : VColor.VColor( rgb = (255,255,255)),
            ( VPalette.ColorGroup.Disabled, VPalette.ColorRole.ToolTipBase) : VColor.VColor( rgb = (0,0,0)),
            ( VPalette.ColorGroup.Disabled, VPalette.ColorRole.ToolTipText) : VColor.VColor( rgb = (255,255,255)),
            ( VPalette.ColorGroup.Inactive, VPalette.ColorRole.WindowText) : VColor.VColor( rgb = (255,255,255)),
            ( VPalette.ColorGroup.Inactive, VPalette.ColorRole.Button) : VColor.VColor( rgb = (255,255,255)),
            ( VPalette.ColorGroup.Inactive, VPalette.ColorRole.Light) : VColor.VColor( rgb = (255,255,255)),
            ( VPalette.ColorGroup.Inactive, VPalette.ColorRole.Midlight) : VColor.VColor( rgb = (255,255,255)),
            ( VPalette.ColorGroup.Inactive, VPalette.ColorRole.Dark) : VColor.VColor( rgb = (255,255,255)),
            ( VPalette.ColorGroup.Inactive, VPalette.ColorRole.Mid) : VColor.VColor( rgb = (255,255,255)),
            ( VPalette.ColorGroup.Inactive, VPalette.ColorRole.Text) : VColor.VColor( rgb = (255,255,255)),
            ( VPalette.ColorGroup.Inactive, VPalette.ColorRole.BrightText) : VColor.VColor( rgb = (255,255,255)),
            ( VPalette.ColorGroup.Inactive, VPalette.ColorRole.ButtonText) : VColor.VColor( rgb = (255,255,255)),
            ( VPalette.ColorGroup.Inactive, VPalette.ColorRole.Base) : VColor.VColor( rgb = (0,0,0)),
            ( VPalette.ColorGroup.Inactive, VPalette.ColorRole.Window) : None,
            ( VPalette.ColorGroup.Inactive, VPalette.ColorRole.Shadow) : VColor.VColor( rgb = (0,0,0)),
            ( VPalette.ColorGroup.Inactive, VPalette.ColorRole.Highlight) : VColor.VColor( rgb = (255,255,255)),
            ( VPalette.ColorGroup.Inactive, VPalette.ColorRole.HighlightedText) : VColor.VColor( rgb = (0,0,0)),
            ( VPalette.ColorGroup.Inactive, VPalette.ColorRole.Link) : VColor.VColor( rgb = (255,255,255)),
            ( VPalette.ColorGroup.Inactive, VPalette.ColorRole.LinkVisited) : VColor.VColor( rgb = (255,255,255)),
            ( VPalette.ColorGroup.Inactive, VPalette.ColorRole.AlternateBase) : VColor.VColor( rgb = (255,255,255)),
            ( VPalette.ColorGroup.Inactive, VPalette.ColorRole.NoRole) : VColor.VColor( rgb = (255,255,255)),
            ( VPalette.ColorGroup.Inactive, VPalette.ColorRole.ToolTipBase) : VColor.VColor( rgb = (0,0,0)),
            ( VPalette.ColorGroup.Inactive, VPalette.ColorRole.ToolTipText) : VColor.VColor( rgb = (255,255,255))
            }

    def copy(self):
        return copy.deepcopy(self)




from ..VWidget import VWidget
from ..VPalette import VPalette
from ..VPainter import VPainter

class VDialog(VWidget):
    def __init__(self, parent=None):
        super(VDialog, self).__init__(parent)
        self._title = None

    def paintEvent(self, event):
        if self.isEnabled():
            if self.isActive():
                color_group = VPalette.ColorGroup.Active
            else:
                color_group = VPalette.ColorGroup.Inactive
        else:
            color_group = VPalette.ColorGroup.Disabled

        fg, bg = self.colors(color_group)
        w, h = self.size()
        if self._title:

            #0123456789012
            #+-| hello |-+
            dash_length = (w -                  # total width of the dialog
                           2 -                  # space for the angles
                           len(self._title) -   # the space for the title itself
                           2 -                  # the two empty spaces on the sides of the title
                           2)                   # the vertical bars
            header = '+' + \
                     "-"*(dash_length/2) + \
                     "| " + \
                     self._title + \
                     " |" + \
                     "-"*(dash_length-(dash_length/2)) + \
                     "+"
        else:
            header = '+'+"-"*(w-2)+"+"

        painter = VPainter(self)
        painter.write(0, 0, header, fg, bg)

        for i in range(0, h-2):
            painter.write(0, i+1, '|'+' '*(len(header)-2)+"|", fg, bg)
        painter.write(0, h-1, '+'+"-"*(len(header)-2)+"+", fg, bg)

    def setTitle(self, title):
        self._title = title

    def minimumSize(self):
        if self._title:
            return (len(self._title) + 8, 2)
        else:
            return (2,2)

    def contentsMargins(self):
        return (1,1,1,1)



from .BufferCommand import BufferCommand
from .CommandResult import CommandResult

class NewLineAfterCommand(BufferCommand):
    def execute(self):
        document = self._document
        cursor = self._cursor

        if self.savedCursorPos() is None:
            self.saveCursorPos()
        
        pos = self.savedCursorPos()

        current_text = document.lineText(pos[0])
        current_indent = len(current_text) - len(current_text.lstrip(' '))

        document.newLineAfter(pos[0])
        document.insertChars( (pos[0]+1, 1), ' '*current_indent )
        document.lineMetaInfo("Change").setData("added", pos[0]+1)
        cursor.toPos((pos[0]+1, 1))
        cursor.toLineEnd()
        return CommandResult(True, None)

    def undo(self):
        self._document.deleteLine(self.savedCursorPos()[0]+1)
        self.restoreCursorPos()




from vaitk import gui
from .Editor import Editor
from . import models
import random
import os


class EditorApp(gui.VApplication):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # We keep them at the App level because the app will be responsible
        # for coordinating the async system in the future.
        self._global_model = models.GlobalState()
        self._buffer_list = models.BufferList()

        self._editor = Editor(self, self._global_model, self._buffer_list)

        self._editor.show()

    def openFile(self, path):
        self._editor.controller.openFile(path)

    def dumpBuffers(self, destination_dir=None):
        """
        Dump the buffers to your home directory in case of a crash.
        Returns the list of files dumped down.
        """

        if destination_dir is None:
            destination_dir = os.path.expanduser("~")

        file_list = []

        for buffer in self._buffer_list.buffers:
            document_text = buffer.document.documentText()
            document_name = buffer.document.filename() or "noname"
            random_number = random.randint(1, 100000)
            path = os.path.join(destination_dir, "vaidump-%s-%d.txt" % (os.path.basename(document_name), random_number))

            with open(path, "w") as f:
                f.write(document_text)

            file_list.append(path)

        return file_list

    @property
    def editor(self):
        return self._editor




from .BufferCommand import BufferCommand
from .CommandResult import CommandResult

class DedentCommand(BufferCommand):
    """
    Dedent the line at the current cursor position.
    """
    def __init__(self, buffer):
        super().__init__(buffer)

    def execute(self):
        cursor = self._cursor
        document = self._document

        if self.savedCursorPos() is None:
            self.saveCursorPos()

        pos = self.savedCursorPos()

        text = document.lineText(pos[0])
        if text[0:4] != '    ':
            return CommandResult(success=False, info=None)

        line_meta = document.lineMetaInfo("Change")
        changed = line_meta.data(pos[0])

        self.saveLineMemento(pos[0], BufferCommand.MEMENTO_REPLACE)

        # Naive implementation. Needs improvement.
        document.deleteChars((pos[0], 1), 4) 
        if changed is None:
            line_meta.setData("modified", pos[0])

        cursor.toPos((pos[0], pos[1]-4))

        return CommandResult(success=True, info=None)

